From iris.algebra Require Import gmap frac deprecated auth gset.
From iris.base_logic Require Import big_op wsat invariants.
From iris.proofmode Require Import tactics.
From iris.heap_lang Require Import proofmode notation.
From stdpp Require Import mapset.
From fletch Require Import makers dequeue enqueue wp_helpers helpers definitions monoids invariants atomic program.

Import dec_agree.

Global Opaque unSOME queue_head queue_tail queue_sent qref next prev 
       obtainLockEnq obtainLockDeq enqueue dequeue makeQueue 
       makeProc tryDequeueEntry.

Section client_concurrent.
  
  Context {Σ : gFunctors} `{!heapG Σ, !fletchG Σ}
          (N : namespace).
    
  Local Notation iProp := (iProp Σ).

(* CLIENT CONCURRENT PROGRAM
===================================================*)

  Notation n_inv_queue q := (N .@ "client" .@ q).

  Definition doWork : val :=
    λ: "pval",
    "pval" <- !"pval".

  Definition scheduler : val :=
    rec: "loop" "h" "t" "s" :=
      let: "p" := dequeue "h" "t" "s" #() in
      match: "p" with
        NONE => #()
      | SOME "p'" => "p'" ;;
        doWork (pval "p'") ;;
        enqueue "h" "t" "s" "p'"
      end ;;
      "loop" "h" "t" "s".

  Definition enqueuer : val :=
    rec: "loop" "h" "t" "s" :=
      let: "p" := makeProc #1 in
      enqueue "h" "t" "s" "p" ;;
       "loop" "h" "t" "s".

  Definition concurrent_client : val :=
    λ: <>,
       let: "q1" := makeQueue #() in
       let: "q1h" := queue_head "q1" in
       let: "q1t" := queue_tail "q1" in
       let: "q1s" := queue_sent "q1" in
       Fork (scheduler "q1h" "q1t" "q1s") ;;
       Fork (scheduler "q1h" "q1t" "q1s") ;;
       Fork (enqueuer "q1h" "q1t" "q1s").

  (* SPECIFICATION AND HELPERS *)
  Definition client_queue γp γq (q : queueAddrT) γ γ' (l : list procAddrT) :=
    (queue N γp γq (q : queueAddrT) γ γ' l ∗
          [∗ list] p ∈ l, ∃ v, pvall p ↦{½} v )%I.

  Definition client_queue_inv γp γq (q : queueAddrT) γ γ' :=
    (inv (n_inv_queue q) (∃ l, client_queue γp γq q γ γ' l))%I.

  Global Instance client_queue_inv_persistent :
    PersistentP (client_queue_inv γp γq q γ γ').
  Proof. apply _. Qed.

  Lemma doWork_spec γp γq p v φ : 
    procs_inv N γp ∗
    qproc N γp γq p ∗
    (pvall p) ↦{½} v ∗
    ((∃ v', qproc N γp γq p ∗ (pvall p) ↦{½} v') -∗ φ #()) 
     ⊢ (WP doWork #(pvall p) {{ φ }})%I.
  Proof.
    iIntros "(#Hprocs & Hproc & Hpval & Hpost)".
    iDestruct "Hproc" as (v') "(HProc & #Hprocinv)".
    wp_let.
    wp_load.
    iInv (n_inv_procs N) as (l) ">(Hauth & Hmap)" "Hclose".
    iDestruct (proc_own_contains with "[$Hauth $HProc]") as %[].
    iDestruct (big_sepM_lookup_acc _ l p with "[Hmap]") as "(Hpval' & Hacc)".
    done. iFrame. iSpecialize "Hclose".
    iDestruct "Hpval'" as (v'') "(Hproc & Hpval')".
    iDestruct (mapsto_agree (pvall p) with "Hpval Hpval'") as %Heq; subst.
    iCombine "Hpval Hpval'" as "Hpval".
    wp_store.
    iDestruct "Hpval" as "(Hpval & Hpval')".
    iCombine "Hproc Hpval'" as "Hproc".
    iSpecialize "Hacc".
    iDestruct ("Hacc" with "[Hproc]") as "Hmap".
    { iExists v''. iFrame. }
    iMod ("Hclose" with "[Hmap Hauth]"). iModIntro.
    iExists l. iFrame. iModIntro.
    iApply "Hpost". iExists (pvalv v''). iFrame.
    iExists v'. iFrame; iFrame "#".
  Qed.

  Lemma scheduler_spec γp γq q γ γ' :
    (client_queue_inv γp γq q γ γ' ∗ procs_inv N γp
    ⊢ WP (scheduler #(qhead q) #(qtail q) #(qsent q)) {{ _, True }})%I.
  Proof.
    iIntros "(#Hqinv & #Hprocs)".
    iLöb as "IH".
    wp_rec. wp_let. wp_let.
    wp_bind (dequeue _ _ _ _).
    iDestruct (dequeue_spec N γp γq q γ γ' (↑(n_inv_queue q)) with "[]") as
        "Hdeq". iFrame "#".
    iApply wp_wand_r. iSplitL "Hdeq".
    iApply ("Hdeq" $! (True)%I
                   (fun l ret =>
                      ((⌜l = []⌝ ∗ ⌜ret = NONEV⌝) ∨
                       (∃ p l' v, ⌜l = p :: l'⌝ ∗
                        qproc N γp γq p ∗ ▷ (pvall p) ↦{½} v ∗ 
                        ⌜ret = option_procAddrT_to_val (Some p)⌝))%I)
            with "[]"); last done.
    { iIntros "!# _".
      iInv (n_inv_queue q) as (l) "(Hq & Hlist)" "Hclose".
      iMod (fupd_intro_mask' _ ∅) as "Hvs"; first set_solver.
      iModIntro. iExists l. iFrame. iSplit.
      - iIntros "Hq". iMod "Hvs". iMod ("Hclose" with "[Hq Hlist]").
        iNext. iExists l. iFrame. done.
      - iIntros (y) "Hq". destruct l.
        iDestruct "Hq" as "[Hq | Hq]".
        iDestruct "Hq" as (p l) "(% & Hq & Hqproc & %)".
        inversion H. iMod "Hvs".
        iDestruct "Hq" as "(% & Hq & %)".
        iMod ("Hclose" with "[Hq Hlist]").
        iNext. iExists []. iFrame. iModIntro. iLeft. iSplit=>//. 
        iDestruct "Hq" as "[Hq | Hq]".
        iDestruct "Hq" as (p' l') "(% & Hq & Hqproc & %)".
        iDestruct "Hlist" as "(Hp & Hlist)".
        iDestruct "Hp" as (v) "(Hpval & Hp)".
        inversion H; subst.
        iMod "Hvs". iMod ("Hclose" with "[Hq Hlist]").
        iNext. iExists l'. iFrame.
        iModIntro. iRight. iExists p',l',v. iFrame. iSplit=>//.
        iDestruct "Hq" as "(% & Hq & %)". inversion H. }

    iIntros (v) "Hx". iDestruct "Hx" as (l) "[(% & %) | Hsome]"; subst.
    - wp_let. wp_match.  iApply "IH".
    - iDestruct "Hsome" as (p l' v') "(% & Hqproc & Hpval & %)".
      wp_let. rewrite H0 /=. wp_match.
      wp_apply (pval_spec).
      wp_bind (doWork _). iApply (doWork_spec γp γq). iFrame; iFrame "#".
      iIntros "Hx". iDestruct "Hx" as (v0) "(Hpval & HP')". wp_value. wp_seq.
      iDestruct (enqueue_spec N γp γq q p γ γ' (↑(n_inv_queue q)) with "[$Hpval]") as
          "Henq". iFrame "#".
      iApply wp_wand_r. iSplitL. 
      iApply ("Henq" $! (pvall p ↦{½} v0)%I
                     (fun l ret => ⌜ret = #(is_nil l)⌝)%I
              with "[]"); last iFrame=>//. 
      { iIntros "!# Hpval".
        iInv (n_inv_queue q) as (l0) "(Hq & Hlist)" "Hclose".
        iMod (fupd_intro_mask' _ ∅) as "Hvs"; first set_solver.
        iModIntro. iExists l0. iFrame "Hq". iSplit.
        - iIntros "Hq". iMod "Hvs". iMod ("Hclose" with "[Hq Hlist]").
          iNext. iExists l0. iFrame. iModIntro. iFrame.
        - iIntros (y) "(Hq & %)".
          iDestruct (big_sepL_app _ l0 [p] with "[Hlist Hpval]") as "Hlist".
          iDestruct (big_sepL_later with "Hlist") as "Hlist". iFrame. simpl.
          iSplitL=>//. iModIntro. iExists v0. iFrame.
          iMod "Hvs". iMod ("Hclose" with "[Hq Hlist]").
          iDestruct (big_sepL_later with "Hlist") as "Hlist".
          iNext. iExists (l0 ++ [p]). iFrame. done. }
      iIntros. wp_seq. iApply "IH".      
  Qed.

  Lemma enqueuer_spec γp γq q γ γ' :
    (client_queue_inv γp γq q γ γ' ∗ procs_inv N γp 
    ⊢ WP (enqueuer #(qhead q) #(qtail q) #(qsent q)) {{ _, True }})%I.
  Proof.
    iIntros "(#Hqinv & #Hprocs)".
    iLöb as "IH".
    wp_rec. wp_let. wp_let.
    wp_bind (makeProc _).
    iApply wp_wand_l; iSplitL;
      [|iApply (makeProc_spec N γp γq #1 with "[Hprocs]");
        iFrame "#"]; iIntros (v) "H".
    iDestruct "H" as (p) "(% & (Hqproc & Hpval))". subst. simpl. wp_let.
    wp_bind (enqueue _ _ _ _).
    iDestruct (enqueue_spec N γp γq q p γ γ' (↑(n_inv_queue q)) with "[$Hqproc]") as
        "Henq". iFrame "#".
    iApply wp_wand_r. iSplitL.
    iApply ("Henq" $! (pvall p ↦{½} #1)%I
                   (fun l ret => ⌜ret = #(is_nil l)⌝)%I
            with "[]"); last iFrame=>//. 
    { iIntros "!# Hpval".
      iInv (n_inv_queue q) as (l0) "(Hq & Hlist)" "Hclose".
      iMod (fupd_intro_mask' _ ∅) as "Hvs"; first set_solver.
      iModIntro. iExists l0. iFrame "Hq". iSplit.
      - iIntros "Hq". iMod "Hvs". iMod ("Hclose" with "[Hq Hlist]").
        iNext. iExists l0. iFrame. iModIntro. iFrame.
      - iIntros (y) "(Hq & %)".
        iDestruct (big_sepL_app _ l0 [p] with "[Hlist Hpval]") as "Hlist".
        iDestruct (big_sepL_later with "Hlist") as "Hlist". iFrame. simpl.
        iSplitL=>//. iModIntro. iExists #1. iFrame.
        iMod "Hvs". iMod ("Hclose" with "[Hq Hlist]").
        iDestruct (big_sepL_later with "Hlist") as "Hlist".
        iNext. iExists (l0 ++ [p]). iFrame. done. }
    iIntros. wp_seq. iApply "IH".
  Qed.

  Lemma concurrent_client_spec :
    (WP concurrent_client #() {{ v, ⌜v = #()⌝ }})%I.
  Proof.
    wp_let.
    
    (* Make procs_inv *)
    iMod (own_alloc (ProcAuth ∅)) as (γp) "Hpa". done.
    iMod (inv_alloc (n_inv_procs N) ⊤ _ with "[Hpa]") as "#Hprocsinv". 
    iNext. iExists ∅. iFrame. by rewrite big_sepM_empty.
    
    (* Make queues_inv *)
    iMod (own_alloc (LinkAll ∅)) as (γq) "Hqa". done.
    iMod (inv_alloc (n_inv_queues N) ⊤ _ with "[Hqa]") as "#Hqueuesinv". 
    iNext. iExists ∅. iFrame. by rewrite big_sepM_empty.

    (* let: "q1" := makeQueue #() in *)
    wp_bind (makeQueue _).
    iApply wp_wand_l; iSplitL; [|iApply (makeQueue_spec N γp γq with "[Hqueuesinv]"); iFrame "#"]; iIntros (q1) "Hq1".
    iDestruct "Hq1" as (γlist1 γlink1 w1) "(% & Hq1)". wp_let.
    (* let: "q1h" := queue_head "q1" in *)
    rewrite H.
    wp_apply queue_head_spec. wp_let.    
    (* let: "q1t" := queue_tail "q1" in *)
    wp_apply queue_tail_spec. wp_let.
    (* let: "q1s" := queue_sent "q1" in *)
    wp_apply queue_sent_spec. wp_let.

    set (Queue l := (queue N γp γq w1 γlist1 γlink1 l)).

    (* Make client_queues_inv q1 *)
    iMod (inv_alloc (n_inv_queue w1) ⊤ _ with "[Hq1]") as "#Hqclient".
    iModIntro. iExists []. iFrame. auto.

    wp_bind (Fork _). iApply wp_fork. iSplitL. iModIntro. wp_seq.
    wp_bind (Fork _). iApply wp_fork. iSplitL. iModIntro. wp_seq.
    (* wp_bind (Fork _). iApply wp_fork. iSplitL. iModIntro. wp_seq. *)
    iApply wp_fork. iSplitL. iModIntro. auto.

    iDestruct (enqueuer_spec _ _ w1 with "[]") as "H"; iFrame "#"; iModIntro; iFrame.
    iDestruct (scheduler_spec _ _ w1 with "[]") as "H"; iFrame "#"; iModIntro; iFrame.
    iDestruct (scheduler_spec _ _ w1 with "[]") as "H"; iFrame "#"; iModIntro; iFrame.
  Qed.

End client_concurrent.
