From iris.algebra Require Import gmap frac deprecated auth gset.
From iris.base_logic Require Import big_op wsat invariants.
From iris.proofmode Require Import tactics.
From iris.heap_lang Require Import proofmode notation.
From stdpp Require Import mapset.
From fletch Require Import makers dequeue enqueue wp_helpers helpers definitions monoids invariants atomic program.

Import dec_agree.

Global Opaque unSOME queue_head queue_tail queue_sent qref next prev 
       obtainLockEnq obtainLockDeq enqueue dequeue makeQueue 
       makeProc tryDequeueEntry.

Section client_sequential.
  
  Context {Σ : gFunctors} `{!heapG Σ, !fletchG Σ}
          (N : namespace).
    
  Local Notation iProp := (iProp Σ).

(* CLIENT SEQUENTIAL PROGRAM
===================================================*)

  Definition testSequentialQueue : val :=
    λ: <>,
       let: "p" := makeProc #42 in
       let: "q1" := makeQueue #() in
       let: "q1h" := queue_head "q1" in
       let: "q1t" := queue_tail "q1" in
       let: "q1s" := queue_sent "q1" in
       let: "q2" := makeQueue #() in
       let: "q2h" := queue_head "q2" in
       let: "q2t" := queue_tail "q2" in
       let: "q2s" := queue_sent "q2" in
         (enqueue "q1h" "q1t" "q1s" "p") ;;
         let: "p'" := unSOME (dequeue "q1h" "q1t" "q1s" #()) in
           (enqueue "q2h" "q2t" "q2s" "p'") ;;
           (tryDequeueEntry "q2h" "q2t" "q2s" "p'").  

  Lemma testSequentialQueue_spec : 
    (WP testSequentialQueue #() {{ v, ⌜v = #true⌝ }})%I.
  Proof.
    wp_let.
    
    (* Make procs_inv *)
    iMod (own_alloc (ProcAuth ∅)) as (γp) "Hpa". done.
    iMod (inv_alloc (n_inv_procs N) ⊤ _ with "[Hpa]") as "#Hprocsinv". 
    iNext. iExists ∅. iFrame. by rewrite big_sepM_empty.
    
    (* Make queues_inv *)
    iMod (own_alloc (LinkAll ∅)) as (γq) "Hqa". done.
    iMod (inv_alloc (n_inv_queues N) ⊤ _ with "[Hqa]") as "#Hqueuesinv". 
    iNext. iExists ∅. iFrame. by rewrite big_sepM_empty.

    (* let: "p" := makeProc #42 in *)
    wp_bind (makeProc _).
    iApply wp_wand_l; iSplitL;
      [|iApply (makeProc_spec N γp γq #42 with "[Hprocsinv]");
        iFrame "#"]; iIntros (v) "H".
    iDestruct "H" as (p) "(% & Hqproc & Hpval)".
    iDestruct "Hqproc" as (v') "(Hpg & #Hprocinv)". wp_let.

    (* let: "q1" := makeQueue #() in *)
    wp_bind (makeQueue _).
    iApply wp_wand_l; iSplitL; [|iApply (makeQueue_spec N γp γq with "[Hqueuesinv]"); iFrame "#"]; iIntros (q1) "Hq1".
    iDestruct "Hq1" as (γlist1 γlink1 w1) "(% & Hq1)". wp_let.
    (* let: "q1h" := queue_head "q1" in *)
    rewrite H0.
    wp_apply queue_head_spec. wp_let.    
    (* let: "q1t" := queue_tail "q1" in *)
    wp_apply queue_tail_spec. wp_let.
    (* let: "q1s" := queue_sent "q1" in *)
    wp_apply queue_sent_spec. wp_let.

    (* Let: "q2" := makeQueue #() in *)
    wp_bind (makeQueue _).
    iApply wp_wand_l; iSplitL;
      [|iApply (makeQueue_spec N γp γq with "[Hqueuesinv]");
        iFrame "#"]; iIntros (q2) "Hq2".
    iDestruct "Hq2" as (γlist2 γlink2 w2) "(% & Hq2)". wp_let.
    (* let: "q2h" := queue_head "q2" in *)
    rewrite H1.
    wp_apply queue_head_spec. wp_let.    
    (* let: "q2t" := queue_tail "q2" in *)
    wp_apply queue_tail_spec. wp_let.
    (* let: "q2s" := queue_sent "q2" in *)
    wp_apply queue_sent_spec. wp_let.

    set (Queue1 l := (queue N γp γq w1 γlist1 γlink1 l)).
    set (Queue2 l := (queue N γp γq w2 γlist2 γlink2 l)).
    
    (* (enqueue "q1h" "q1t" "q1s" "p") *)
    wp_bind (enqueue _ _ _ _). rewrite H.
    iDestruct (enqueue_spec N γp γq _ _ _ _ _ with "[Hpg]")
      as "Henq". iFrame "#". iExists v'. iFrame.
    iApply wp_wand_r; iSplitL "Henq Hq1".
    iApply ("Henq" $! (▷ Queue1 [])%I
                   (fun l ret => (Queue1 [p])) with "[]").
    { iIntros "!# Hq1".
      iModIntro. iExists []. iFrame. iSplit.
      iIntros "H1". iModIntro. iFrame. 
      iIntros (y) "H1".
      iModIntro. simpl. iDestruct "H1" as "(H1 & _)". iFrame. }
    iFrame. simpl.
    iIntros (v0) "Hq1". iDestruct "Hq1" as (?) "Hq1". wp_seq. 

    (* let: "p'" := unSOME (dequeue "q1h" "q1t" "q1s" #()) in *)
    wp_bind (dequeue _ _ _ _).
    iDestruct (dequeue_spec N γp γq w1 _ _ ⊤ with "[]") as
        "Hdeq". done. iFrame; iFrame "#".
    iApply wp_wand_r; iSplitL "Hdeq Hq1".
    iApply ("Hdeq" $! (▷ Queue1 [p])%I
                   (fun l ret => (Queue1 [] ∗ qproc N γp γq p ∗
                                 ⌜ret = option_procAddrT_to_val (Some p)⌝)%I)
            with "[]").
    { iIntros "!# Hq1".
      iMod (fupd_intro_mask' ⊤ ∅) as "Hvs"; first set_solver.
      iModIntro. iExists [p]. iFrame. iSplit.
      - iIntros "Hq1". iFrame. 
      - iMod "Hvs". iIntros (y) "[Hqdeq | Hqempt]".
        + iModIntro.
          iDestruct "Hqdeq" as (p0 l') "(% & Hq1 & Hproc1)".
          inversion H3; subst. iFrame.
        + iDestruct "Hqempt" as "(% & _)". inversion H3. }
    iFrame.
    iIntros (v1) "Hq1". iDestruct "Hq1" as (?) "(Hq1 & Hp & %)".
    rewrite H4. wp_apply (unSOME_spec). wp_let.

    (* enqueue "q2h" "q2t" "q2s" "p'" *)
    wp_bind (enqueue _ _ _ _). 
    iDestruct (enqueue_spec N γp γq _ _ _ _ _ with "[Hp]")
      as "Henq". iFrame; iFrame "#".
    iApply wp_wand_r; iSplitL "Henq Hq2".
    iApply ("Henq" $! (▷ Queue2 [])%I
                   (fun l ret => (Queue2 [p])) with "[]").
    { iIntros "!# Hq1".
      iModIntro. iExists []. iFrame. iSplit.
      - iIntros "H1". iFrame. done.
      - iIntros (y) "H1". iModIntro. iDestruct "H1" as "(H1 & _)". iFrame. }
    iFrame. simpl.
    iIntros (v2) "Hq2". iDestruct "Hq2" as (?) "Hq2". wp_seq. 

    (* tryDequeueEntry "q2h" "q2t" "q2s" "p'" *)
    iDestruct (tryDequeueEntry_spec N γp γq w2 p _ _ ⊤ with "[]") as
        "HdeqEntry". iFrame; iFrame "#".
    iApply wp_wand_r; iSplitL "HdeqEntry Hq2".
    iApply ("HdeqEntry" $! (▷ Queue2 [p])%I
                        (fun l ret => (Queue2 [] ∗ qproc N γp γq p ∗
                                       ⌜ret = #true⌝)%I) with "[]").
    { iIntros "!# Hq2".
      iMod (fupd_intro_mask' ⊤ (↑n_inv_proc N p)) as "Hvs"; first set_solver.
      iModIntro. iExists [p]. iFrame. iSplit.
      - iIntros "Hq1". iFrame.
      - iMod "Hvs". iIntros (y) "[Hqlock | Hqempt]".
        + iModIntro.
          iDestruct "Hqlock" as "[(% & Hdeq) | (% & Habs)]".
          * iDestruct "Hdeq" as (l1 l2) "(% & Hq1 & Hp)".
            have->:(l1 ++ l2 = []) by apply (singleton_app_cons p). iFrame.
          * destruct H6. constructor.
        + iDestruct "Hqempt" as "(% & _)". inversion H6. }
    iFrame. iIntros (v3) "Hq2". iDestruct "Hq2" as (?) "(Hq2 & Hp & %)".
    done.
   Qed.

End client_sequential.

