From iris.base_logic Require Import big_op wsat invariants.
From iris.heap_lang Require Import notation.
From iris.algebra Require Import gmap frac agree auth gset.
From stdpp Require Import mapset.

Notation n_inv_queues N := (N .@ "queues").
Notation n_inv_procs N := (N .@ "procs").
Notation n_inv_proc N p := (N .@ "proc" .@ p).

Notation "½" := (1/2)%Qp.
Notation "¼" := (1/4)%Qp.
Notation "¾" := (1/4 + 1/4 + 1/4)%Qp.

Record queueAddrT := QueueAddrT {
  qhead : loc; qtail : loc; qsent : loc
}.

Instance queueAddrT_eq_dec : EqDecision queueAddrT.
Proof. solve_decision. Defined.
Instance queueAddrT_countable : Countable queueAddrT.
Proof.
  set (to_prod a := (qhead a, qtail a, qsent a)).
  set (of_prod p := Some (QueueAddrT (fst (fst p)) (snd (fst p)) (snd p))).
  apply (inj_countable to_prod of_prod).
  by intros [].
Defined.

Record procAddrT := ProcAddrT {
   pvall : loc; pqueuel : loc; pprevl : loc; pnextl : loc
}.
Instance procAddrT_eq_dec : EqDecision procAddrT.
Proof. solve_decision. Defined.
Instance procAddrT_countable : Countable procAddrT.
Proof.
  set (to_prod a := (pvall a, pqueuel a, pprevl a, pnextl a)).
  set (of_prod p := Some (ProcAddrT (fst (fst (fst p))) (snd (fst (fst p))) (snd (fst p)) (snd p))).
  apply (inj_countable to_prod of_prod).
  by intros [].
Defined.

Record procValT := ProcValT' {
   pvalv : val;
   pqueuev : option queueAddrT;
   pprevv : option procAddrT;
   pnextv : option procAddrT
}.
Instance procValT_eq_dec : EqDecision procValT.
Proof. solve_decision. Defined.

Definition ProcValT v q p n : procValT :=
  {| pvalv := v; pqueuev := q; pprevv := p; pnextv := n |}.

Definition of_list' := of_list (C := gset procAddrT).

Definition is_nil {A} (l : list A) : bool := match l with [] => true | _ => false end. 

(* CONVERSIONS *)
Definition procAddrT_to_val (pv : procAddrT) : val :=
  (#(pvall pv), (#(pqueuel pv),(#(pprevl pv), #(pnextl pv)))).

Definition option_procAddrT_to_val (pv : option procAddrT) : val :=
  match pv with
    | None => NONEV
    | Some v => SOMEV (procAddrT_to_val v)
  end.

Lemma option_procAddrT_to_val_somev : ∀ x,
  option_procAddrT_to_val (Some x) = SOMEV (procAddrT_to_val x).
Proof. auto. Qed.  

Lemma option_procAddrT_to_val_some :
  ∀ x, SOME (procAddrT_to_val x) = option_procAddrT_to_val (Some x).
Proof. auto. Qed.

Definition queueAddrT_to_val (qv : queueAddrT) : val :=
  (#(qhead qv),(#(qtail qv), #(qsent qv))).

Definition option_queueAddrT_to_val (qv : option queueAddrT) : val :=
  match qv with
    | None => NONEV
    | Some q => SOMEV (queueAddrT_to_val q)
  end.

Instance procAddrT_to_val_inj : Inj (=) (=) procAddrT_to_val.
Proof.
  rewrite /procAddrT_to_val.
  intros [] [] H.
    by inversion H.
Qed.

Lemma somev_inj :
  ∀ p q, SOMEV p = SOMEV q -> p = q.
Proof. intros. by inversion H.
Qed.

Instance option_procAddrT_to_val_inj : Inj (=) (=) option_procAddrT_to_val.
Proof.
  intros [] [] H0.
  - rewrite /option_procAddrT_to_val in H0.
    apply somev_inj in H0. by simplify_eq.
  - simplify_eq.
  - simplify_eq.
  - done.
Qed.

Instance queueAddrT_to_val_inj : Inj (=) (=) queueAddrT_to_val.
Proof. intros [] [] ?; by simplify_eq.
Qed.

Lemma option_queueAddrT_to_val_eq :
  ∀ q q', option_queueAddrT_to_val q = option_queueAddrT_to_val q' → q = q'.
Proof. 
  intros [] [] H0.
  - rewrite /option_queueAddrT_to_val in H0.
    apply somev_inj in H0. by simplify_eq.
  - simplify_eq.
  - simplify_eq.
  - done.
Qed.

Lemma not_elem_of_of_list' (x : procAddrT) (l : list procAddrT) :
  x ∉ of_list' l ↔ x ∉ l.
Proof. by rewrite elem_of_of_list. Qed.

Lemma of_list'_union_tail p l :
  {[ p ]} ∪ of_list' l = of_list' (l ++ [p]).
Proof. set_solver. Qed.

Lemma of_list'_middle p l1 l2 :
  of_list' (l1 ++ p :: l2) =
  {[p]} ∪ (of_list' (l1 ++ l2)).
Proof. set_solver. Qed.

Lemma is_nil_app (e : procAddrT) l :
  is_nil (l ++ [e]) = false.
Proof.
  destruct l; compute; done.
Qed.

Opaque procAddrT_to_val option_procAddrT_to_val
       queueAddrT_to_val option_queueAddrT_to_val.
