From iris.algebra Require Import gmap frac deprecated auth gset.
From iris.base_logic Require Import big_op wsat invariants.
From iris.proofmode Require Import tactics.
From iris.heap_lang Require Import proofmode notation.
From stdpp Require Import mapset.
From fletch Require Import wp_helpers helpers definitions monoids invariants
     atomic program.

Import dec_agree.

Section dequeue.

  Arguments queue_dll : simpl never.

  Context {Σ : gFunctors} `{!heapG Σ, !fletchG Σ}
          (N : namespace) (γp γq : gname).
    
  Local Notation iProp := (iProp Σ).
  
  Definition deq_α q γ γ' := (fun l => ▷ queue N γp γq q γ γ' l)%I.
  Definition deq_β q γ γ' QP (nilret : val) :=
    (fun (l : list procAddrT) ret => (QP l ret)
            ∨ (⌜l = []⌝ ∗ queue N γp γq q γ γ' [] ∗ ⌜ret = nilret⌝))%I.
  
  Definition obtainLockDeq_atomic_spec (q : queueAddrT) γ γ' P Q QP Ei Eo nilret :=
    atomic_shift (deq_α q γ γ') (deq_β q γ γ' QP nilret) Ei Eo P Q.

  Lemma obtainLockDeq_spec (q : queueAddrT) γ γ' (h : loc) P Q QP Ei Eo nilret φ :
    obtainLockDeq_atomic_spec q γ γ' P Q QP Ei Eo nilret ∗ 
    (∃ hv, h ↦ hv ∗ ⌜hv ≠ NONEV⌝) ∗ P ∗
    (∀ r hv', h ↦ (option_procAddrT_to_val hv') ∗
        ((⌜r = #true⌝ ∗ ⌜hv' = None⌝ ∗ Q [] nilret) ∨ 
         (∃ tv hv'' l, ⌜r = #false⌝ ∗ ⌜hv' = Some hv''⌝ ∗ P ∗
          queue_lock N γp γq q γ γ' (SOMEV #(qsent q)) hv' tv l)) -∗ φ r)
     ⊢ WP obtainLockDeq #(qhead q) #(qsent q) #h {{ φ }}.
  Proof.
    iIntros "(#Hvs & Hh & HP & Hret)".
    iLöb as "IH".
    iDestruct "Hh" as (hv) "(Hh & %)".
    destruct q; simpl.
    wp_rec. wp_let. wp_let. wp_load. wp_let.
    wp_op=>[->|?]. wp_if.
    - (* h is SENTINEL *)
      wp_if.
      wp_bind (!_)%E.
      iApply (wp_mask_mono Eo); first done.
      iMod ("Hvs" with "HP") as (l) "[Hq HPvs]".
      iDestruct "Hq" as (hv0 hvOP tvOP) "(Hqhead & Hlist & Hrest)". simpl.
      wp_load.
      destruct (decide (hv0 = NONEV)) as [->|Hneq].
      + (* we have read NONE, get Q *)
        iDestruct "Hrest" as "(Hlink & Hlist'' & [HsentEq | (HsentEq & Hhead' & Hlist' & Htail & Hqa & #Hprocs & Hdll)])"; iDestruct "HsentEq" as %HsentEq.
        { inversion HsentEq. }
        destruct hvOP. inversion HsentEq. simpl.
        iDestruct "HPvs" as "[_ HQvs]".
        iDestruct (queue_dll_none with "Hdll") as %->.
        iDestruct ("HQvs" $! nilret) as "HQvs".
        iMod ("HQvs" with "[-Hh Hret]") as "HQ".
        iRight. iSplitR=>//. iSplitL.
        iExists NONEV, None, tvOP. iFrame. iRight. iFrame. eauto. done.
        iModIntro. wp_store.
        wp_load. wp_op=>[Heq|Hneq]; wp_if.
        iApply ("Hret" $! #true None).
        iFrame. iLeft. iFrame. auto.
        { destruct Hneq=>//. }
      + (* We have read some value, return P *)
        iDestruct "HPvs" as "[HPvs _]".
        iDestruct "Hrest" as "(Hlink & Hlist'' & [Hveq | (Hveq & Hhead' & Hlist' & Htail & Hqa & #Hprocs & Hdll)])"; iDestruct "Hveq" as %Hveq.
        * iMod ("HPvs" with "[-Hh Hret]") as "HP".
          iFrame. iExists hv0,hvOP,tvOP. iFrame.
          iLeft. done.
          iModIntro.
          wp_store.
          wp_load. wp_op=>[Heq|Hneq']; wp_if.
          { contradiction. }
          iApply ("IH" with "[Hh] HP [Hret]").
          iExists hv0. iFrame. auto. iFrame.
        * iMod ("HPvs" with "[-Hh Hret]") as "HP".
          iFrame. iExists hv0,hvOP,tvOP. iFrame.
          iRight. iFrame. auto. iModIntro.
          wp_store.
          wp_load. wp_op=>[Heq|Hneq']; wp_if.
          { contradiction. }
          iApply ("IH" with "[Hh] HP [Hret]").
          iExists hv0. iFrame. auto. iFrame.
    - (* h do not point to sentinel, try CAS *)
      wp_if. wp_bind (CAS _ _ _)%E.
      iApply (wp_mask_mono Eo); first done.
      iMod ("Hvs" with "HP") as (l) "[Hq [HPvs _]]".
      iDestruct "Hq" as (hv0 hvOP tvOP) "(>Hqhead & Hsent & Hlink & Hlist & Hrest)".
      simpl. destruct (decide (hv = hv0)) as [->|Hneq].
      + (* CAS will succeed, since hv = hv0  *)
        iDestruct "Hrest" as "[>Hsent' | [>HvOp [>Hqhead' Hlock]]]"=>//.
        iDestruct "Hsent'" as %Hsent'=>//.
        iDestruct "HvOp" as %HvOp. simpl.
        iCombine "Hqhead" "Hqhead'" as "Hqhead".
        wp_cas_suc.
        iDestruct "Hqhead" as "[Hqhead Hqhead']".
        iMod ("HPvs" with "[Hqhead Hsent Hlink Hlist]") as "HP".
        iExists (SOMEV #qsent),hvOP,tvOP. iFrame. iLeft. auto.
        iModIntro. wp_op. wp_if.
        iApply ("Hret" $! #false hvOP with "[-]"). subst. iFrame. iRight.
        destruct hvOP. iExists tvOP,p,l. iFrame. eauto.
        { rewrite /option_procAddrT_to_val in H. destruct H=>//. }
      + (* CAS will not succeed *)
        wp_cas_fail.
        iMod ("HPvs" with "[Hqhead Hlist Hsent Hlink Hrest]") as "HP".
        iExists hv0,hvOP,tvOP. iFrame.
        iModIntro. wp_op.
        wp_if. wp_bind (!_)%E.
        iApply (wp_mask_mono Eo); first done.
        iMod ("Hvs" with "HP") as (l') "[Hq HPvs]".
        iDestruct "Hq" as (hv0' hvOP' tvOP') "(Hqhead & Hlist & Hrest)". simpl.
        wp_load.
        destruct (decide (hv0' = NONEV)) as [->|Hneq'].
        * (* we have read NONE, get Q *)
          iDestruct "Hrest" as "(Hlink & Hlist'' & [Hseq | (Hseq & Hhead' & Hlist' & Htail & Hqa & #Hprocs & Hdll)])"; iDestruct "Hseq" as %Hseq.

          { inversion Hseq. }
          destruct hvOP'. inversion Hseq. simpl.
          iDestruct "HPvs" as "[_ HQvs]".
          iDestruct (queue_dll_none with "Hdll") as %->.
          iDestruct ("HQvs" $! nilret) as "HQvs".
          iMod ("HQvs" with "[-Hh Hret]") as "HQ".
          iRight. iSplitR=>//. iSplitL.
          iExists NONEV, None, tvOP'. iFrame. iRight. iFrame. eauto. done.
          iModIntro. wp_store.
          wp_load. wp_op=>[Heq|Hneq']; wp_if.
          iApply ("Hret" $! #true None with "[-]"). 
          iFrame. iLeft. iFrame. auto.
          { destruct Hneq'=>//. }
        *  (* We have read some value, return P *)
          iDestruct "HPvs" as "[HPvs _]".
          iMod ("HPvs" with "[-Hh Hret]") as "HP".
          iFrame. iExists hv0',hvOP',tvOP'. iFrame.
          iModIntro. wp_store.
          wp_load. wp_op=>[Heq|Hneq'']; wp_if.
          { contradiction. }
          iApply ("IH" with "[Hh] HP Hret").
          iExists hv0'. iFrame. auto.
  Qed.

  Lemma dequeue_spec :
    ∀ q γ γ' E,
      procs_inv N γp ⊢
      atomic_wp
      (deq_α q γ γ')
      (deq_β q γ γ' (fun l ret => 
                       (∃ p l', ⌜l = p :: l'⌝ ∗ queue N γp γq q γ γ' l' ∗
                                 qproc N γp γq p ∗
                                 ⌜ret = option_procAddrT_to_val (Some p)⌝)) NONEV)
      ∅ E (dequeue #(qhead q) #(qtail q) #(qsent q) #())%E.
  Proof.
    iIntros (q γ γ' E) "#Hprocsinv".
    iIntros (P Q) "#Hvs HP".
    wp_lam. wp_lam. wp_let.

    (* let: "h" := ref !"head" in *)
    wp_bind (! _)%E.
    iApply (wp_mask_mono E); first done.
    iMod ("Hvs" with "HP") as (l) "[Hq HPvs]".
    iDestruct "Hq" as (hv0 hvOP tvOP) "(Hqhead & Hsent & Hlink & Hlist & Hrest)".
    wp_load.
    (* if: !"h" = NONE then NONE *)
    destruct (decide (hv0 = NONEV)) as [->|Hneq].
    - (* !"h" = NONE *)
      iDestruct "Hrest" as "[Hsenteq | (Hsenteq & Hhead' & Hlist' & Htail & Hqa
            & #Hprocs & Hdll)]"; iDestruct "Hsenteq" as %Hsenteq.
      { inversion Hsenteq. }
      destruct hvOP. inversion Hsenteq. simpl.
      iDestruct "HPvs" as "[_ HQvs]".
      iDestruct (queue_dll_none with "Hdll") as %->.
      iDestruct ("HQvs" $! NONEV) as "HQvs".
      iMod ("HQvs" with "[-]") as "HQ".
      iRight. iSplitR=>//. iSplitL.
      iExists NONEV, None, tvOP. iFrame. iRight. iFrame. eauto. done.
      iModIntro.
      wp_alloc h as "Hh". wp_let. clear tvOP.
      wp_load. wp_op=>[Heq|Hneq]; wp_if.
      iExists []. iFrame.
      { destruct Hneq=>//. }
    - (* !"h" ≠ NONE *)
      iDestruct "HPvs" as "[HPvs _]".
      iMod ("HPvs" with "[-]") as "HP".
      iFrame. iExists hv0,hvOP,tvOP. iFrame.
      iModIntro.
      wp_alloc h as "Hh". wp_let. clear hvOP tvOP l.
      wp_load. wp_op=>[Heq|Hneq']; wp_if.
      { contradiction. }

      (* let: "obtLock" := tryObtainLock "head" "s" "h" in *)
      wp_apply (obtainLockDeq_spec q γ γ' h P Q with "[-]"). 
      iFrame. iFrame "#".
      iSplitL "Hh". iExists hv0=>//. iFrame=>//.
      iIntros (r hv') "(Hh & [(% & % & HQ) | Hrfalse])"; subst. wp_let. wp_if.
      + iExists []=>//.
      + clear Hneq Hneq' hv0.
        iDestruct "Hrfalse" as (tv hv'' l) "(% & % & HP & (Hhead & Hlist & Hqga & Htail & (#Hprocinvs & Hqlist)))"; subst.
        wp_let. wp_if.
        iDestruct (queue_dll_not_nil_head with "Hqlist") as %Hlnotnil. done.
        destruct l. { destruct Hlnotnil=>//. } subst.
        rewrite queue_dll_unfold. 
        iDestruct "Hqlist" as (v0 n0) "(% & Hpg & Hqlist)". 
        inversion H. subst. clear H.
        iDestruct (big_sepL_cons with "Hprocinvs") as "(#Hprocinv & Hprocinvs')".

        (* let: "h'" := unSOME (!"h") in *)
        wp_load. wp_apply unSOME_spec. wp_let.

        (* let: "next" := !(next "h'") in *)
        wp_apply (wp_proc_load_next_ghost N γp γq p). iFrame ; iFrame "#". 
        iIntros (next) "(% & Hpg)". wp_let.

        (* if: "next" = NONE then *)
        wp_op=>[Heqnone|Hneqnone]; wp_if; subst.

        * (* "next" = NONE *)
          destruct n0=>//; subst.
          iDestruct (queue_dll_none with "Hqlist") as %Hlnil; subst.   

          (* "tail" <- NONE *)
          wp_store.

          (* next "h'" <- NONE *)
          wp_apply next_spec.
          wp_bind (_ <- _)%E.
          have->:NONE = option_procAddrT_to_val None. auto.
          iApply (wp_proc_store_next N γp γq _ _ _ _ _ None). 
          iFrame; iFrame "#". iIntros "Hpg". wp_seq.

          (* qref "h'" <- NONE *)
          wp_apply qref_spec.
          wp_apply (wp_proc_store_queue_none). iFrame; iFrame "#".
          iIntros "Hret". iDestruct "Hret" as (γ'') "(Hpg & Hplink & Hinq)". wp_seq.
          
          (* "head" <- "next" *)
          wp_bind (_ <- _)%E.
          iApply (wp_mask_mono E); first done.
          iMod ("Hvs" with "HP") as (l') "[Hq [_ HQvs]]".
          iDestruct "Hq" as (hv0 hvOP tvOP)
                              "(>Hhead' & Hsent & Hlink & >Hlist' & Hrest)".
          iDestruct (list_combine with "[$Hlist $Hlist']") as %Hleq; subst. 
          iDestruct (mapsto_agree (qhead q) with "Hhead Hhead'") as %foo; subst.
          iCombine "Hhead" "Hhead'" as "Hhead".
          wp_store.
          iDestruct (list_update γ ([p]) ([])
                     with "[$Hlist $Hlist']") as ">(Hlist & Hlist')".
          iDestruct "Hhead" as "[Hhead Hhead']".
          iDestruct (link_eq with "[$Hlink $Hplink]") as %Heq; subst.
          iDestruct (inQueue_remove_cons p [] γ'' with "[$Hqga $Hinq]") as ">Hqga".
            by apply disjoint_singleton_l.
          iDestruct ("HQvs" $! (option_procAddrT_to_val (Some p)) with "[-]") 
            as "HQvs". 
          iLeft. iExists p,[]. iFrame. iSplitL ""=>//. iSplitR "Hpg".
          iExists (option_procAddrT_to_val None),None,None. iFrame. iRight. 
          iSplitR=>//. iFrame. iSplitL=>//. iSplitL=>//.
          iExists v0. iFrame; iFrame "#".
          iMod "HQvs". iModIntro. wp_seq. iExists [p]. iFrame.

        * (* "next" ≠ NONE *)
          destruct n0; subst.
          iDestruct (queue_dll_not_nil_head with "Hqlist") as %Hlnotnil'. eauto. 
          iDestruct (queue_dll_do_not_contain with "[$Hpg $Hqlist]") as %Hpnotinl.
          destruct l. destruct Hlnotnil'=>//.
          rewrite queue_dll_unfold.
          iDestruct "Hqlist" as (v1 n1) "(% & Hn1g & Hqlist)". inversion H=>//. 
          iDestruct (big_sepL_cons with "Hprocinvs'") as "(#Hprocinvn & _)".

          (* prev (unSOME "next") <- NONE) *)
          wp_apply unSOME_spec. wp_apply prev_spec.
          wp_apply (wp_proc_store_prev N γp γq _ _ _ _ _ None). iFrame; iFrame "#".
          iIntros "Hng". wp_seq.

          (* next "h'" <- NONE *)
          wp_apply next_spec. wp_apply (wp_proc_store_next N γp γq _ _ _ _ _  None).
          iFrame; iFrame "#". iIntros "Hpg". wp_seq.

          (* qref "h'" <- NONE *)
          wp_apply qref_spec.
          wp_apply (wp_proc_store_queue_none). iFrame; iFrame "#". 
          iIntros "Hret". iDestruct "Hret" as (γ'') "(Hpg & Hplink & Hinq)". wp_seq.

          (* "head" <- "next" *)
          wp_bind (_ <- _)%E.
          iApply (wp_mask_mono E); first done.
          iMod ("Hvs" with "HP") as (l') "[Hq [_ HQvs]]".
          iDestruct "Hq" as (hv0 hvOP tvOP)
                              "(>Hhead' & Hqsent & Hlink & >Hlist' & Hrest)".
          iDestruct (list_combine with "[$Hlist $Hlist']") as %Hleq; subst.
          iDestruct (mapsto_agree (qhead q) with "Hhead Hhead'") as %foo; subst.
          iCombine "Hhead" "Hhead'" as "Hhead".
          wp_store.
          iDestruct (list_update γ (p :: p1 :: l) (p1 :: l)
                     with "[$Hlist $Hlist']") as ">(Hlist & Hlist')".
          iDestruct "Hhead" as "[Hhead Hhead']".
          (* have->: of_list' (p :: p1 :: l) = {[p]} ∪ (of_list' (p1 :: l)) by auto. *)
          iDestruct (link_eq with "[$Hlink $Hplink]") as %Heq; subst.
          iDestruct (inQueue_remove_cons p (p1 :: l)
                     with "[Hqga Hinq]") as ">Hqga".
          { revert Hpnotinl. rewrite -not_elem_of_of_list'. intros.
            by apply disjoint_singleton_l. } iFrame.
          iDestruct ("HQvs" $! (option_procAddrT_to_val (Some p)) with "[-]") 
            as "HQvs".
          iLeft. iExists p. iExists (p1 :: l). 
          iFrame. iSplitL ""=>//. iSplitR "Hpg".
          iExists (option_procAddrT_to_val (Some p1)),(Some p1),tv. iFrame. 
          iRight. iSplitR=>//. iFrame "#"=>//. iFrame. iSplitR=>//.
          iDestruct "Hprocinvs'" as "(_ & Hprocsl)". iFrame "#".
          iExists v1,n1. iFrame=>//. 
          iSplitL. iExists v0. iFrame; iFrame "#". eauto.
          iMod "HQvs". iModIntro. wp_seq.
          iExists (p::p1::l). iFrame.
          { destruct Hneqnone=>//. }
  Qed.
  
  Lemma tryDequeueEntry_spec :
    ∀ q p γ γ' E,
      procs_inv N γp ∗
      proc_inv N γp γq p ⊢ atomic_wp
      (deq_α q γ γ')
      (deq_β q γ γ' (fun l ret => 
                       (⌜p ∈ l⌝ ∗ ∃ l1 l2, ⌜l = l1 ++ p :: l2⌝ ∗ 
                       queue N γp γq q γ γ' (l1 ++ l2) ∗ qproc N γp γq p ∗ 
                       ⌜ret = #true⌝) ∨ (⌜p ∉ l⌝ ∗ queue N γp γq q γ γ' l ∗
                                        ⌜ret = #false⌝)) #false)
      (↑n_inv_proc N p) E
      (tryDequeueEntry #(qhead q) #(qtail q) #(qsent q) (procAddrT_to_val p))%E.
  Proof.
    iIntros (q p γ γ' E) "(#Hprocsinv & #Hpinv)".
    iIntros (P Q) "#Hvs HP".
    wp_lam. wp_lam. wp_let. wp_let.

    (* let: "h" := ref !"head" in *)
    wp_bind (! _)%E.
    iApply (wp_mask_mono E); first done.
    iMod ("Hvs" with "HP") as (l) "[Hq HPvs]".
    iDestruct "Hq" as (hv0 hvOP tvOP) "(Hqhead & Hlist & Hrest)".
    wp_load.
    (* if: !"h" = NONE then NONE *)
    destruct (decide (hv0 = NONEV)) as [->|Hneq].
    - (* !"h" = NONE *)
      iDestruct "Hrest" as "(Hlink & HList & [Hsenteq | (Hsenteq & Hhead' & Hlist' & Htail & Hqa & #Hprocs & Hdll)])"; iDestruct "Hsenteq" as %Hsenteq.
      { inversion Hsenteq. }
      destruct hvOP. inversion Hsenteq. simpl.
      iDestruct "HPvs" as "[_ HQvs]".
      iDestruct (queue_dll_none with "Hdll") as %->.
      iDestruct ("HQvs" $! #false) as "HQvs".
      iMod ("HQvs" with "[-]") as "HQ".
      iRight. iSplitR=>//. iSplitL.
      iExists NONEV, None, tvOP. iFrame. iRight. iFrame. eauto. done.
      iModIntro.
      wp_alloc h as "Hh". wp_let. clear tvOP.
      wp_load. wp_op=>[Heq|Hneq]; wp_if.
      iExists []. iFrame.
      { destruct Hneq=>//. }
    - (* !"h" ≠ NONE *)
      iDestruct "HPvs" as "[HPvs _]".
      iMod ("HPvs" with "[-]") as "HP".
      iFrame. iExists hv0,hvOP,tvOP. iFrame.
      iModIntro.
      wp_alloc h as "Hh". wp_let. clear hvOP tvOP l.
      wp_load. wp_op=>[Heq|Hneq']; wp_if.
      { contradiction. }

      (* let: "obtLock" := tryObtainLock "head" "s" "h" in *)
      wp_apply (obtainLockDeq_spec q γ γ' h P Q with "[-]"). 
      iFrame. iFrame "#". iSplitL "Hh". iExists hv0=>//. iFrame=>//.
      iIntros (r hv') "(Hh & [(% & % & HQ) | Hrfalse])"; subst. wp_let. wp_if.
      + iExists []=>//.
      + clear Hneq Hneq' hv0.
        iDestruct "Hrfalse" as (tv hv'' l) "(% & % & HP & (Hhead & Hlist & Hqga & Htail & (#Hprocinvs & Hqlist)))"; subst.
        wp_let. wp_if.

        (* if: !(qref "p") = SOME ("head",("tail","s")) then *)
        destruct (decide (p ∈ l)) as [Hpinl|Hpnotinl].
        iDestruct (queue_dll_split γp p l q Hpinl with "Hqlist")
          as (v' p' n' l1 l2) "(% & Hpg' & Hqlist1 & Hqlist2)"; subst.

        iDestruct (big_sepL_app _ l1 (p::l2) with "Hprocinvs")
          as "(Hpinvsl1 & Hpinvsl2)".
        iDestruct (big_sepL_cons with "Hpinvsl2") as "(Hpinv' & Hpinvsl2')".
        wp_apply wp_proc_load_qref_ghost. iFrame; iFrame "#".
        iIntros (r) "(% & Hpg)".
        wp_op=>[Heq|Hneq']; wp_if.

        (* let: "next" := !(next "p") in *)
        wp_apply wp_proc_load_next_ghost. iFrame; iFrame "#".
        iIntros (next) "(% & Hpg)". wp_let.

        (* let: "prev" := !(prev "p") in *)
        wp_apply wp_proc_load_prev_ghost. iFrame; iFrame "#".
        iIntros (prev) "(% & Hpg)". wp_let.

        (* (if: "next" = NONE then *)
        wp_op=>[Hnextnone|Hnextnotnone]; wp_if.

        * (* "next" = NONE *)
          subst. destruct n'=>//. 
          iDestruct (queue_dll_none with "Hqlist2") as %->.

          (* "tail" <- "prev" *)
          wp_store.

          (* (if: "prev" = NONE then *)
          wp_op=>[Hprevnone|Hprevnotnone]; wp_if.
          
          { (* "prev" = NONE *)
            subst. destruct p'=>//. destruct l1=>//; simpl.
            
            - (* "h" <- "next" *)
              wp_store.

              (* (prev "p") <- NONE  *)
              wp_apply prev_spec.
              wp_apply (wp_proc_store_prev N γp γq _ _ _ _ _ None). iFrame; iFrame "#".
              iIntros "Hpg". wp_seq.

              (* (next "p") <- NONE  *)
              wp_apply next_spec.
              wp_apply (wp_proc_store_next N γp γq _ _ _ _ _ None). iFrame; iFrame "#".
              iIntros "Hpg". wp_seq.

              (* (qref "p") <- NONE  *)
              wp_apply qref_spec.
              wp_apply (wp_proc_store_queue_none N γp γq). iFrame; iFrame "#".
              iIntros "Hret". iDestruct "Hret" as (γ'') "(Hpg & Hplink & Hinq)".
              wp_seq.

              (* "head" <- !"h" *)
              wp_load.
              wp_bind (_ <- _)%E.
              iApply (wp_mask_mono E); first done.
              iMod ("Hvs" with "HP") as (l') "[Hq [_ HQvs]]".
              iDestruct "Hq" as (hv0 hvOP tvOP)
                                  "(>Hhead' & Hsent & Hlink & >Hlist' & Hrest)".
              iDestruct (list_combine with "[$Hlist $Hlist']") as %Hleq; subst.
              iDestruct (mapsto_agree (qhead q) with "Hhead Hhead'") as %foo; subst.
              iCombine "Hhead" "Hhead'" as "Hhead".
              wp_store.
              iDestruct (link_eq with "[$Hlink $Hplink]") as %Hγeq; subst.
              iDestruct (list_update γ ([p]) ([])
                         with "[$Hlist $Hlist']") as ">(Hlist & Hlist')".
              iDestruct "Hhead" as "[Hhead Hhead']".
              iDestruct (inQueue_remove p [] [] with "[Hqga Hinq]") as ">Hqga".
              { simpl. by apply disjoint_singleton_l. } simpl. iFrame.
              iDestruct ("HQvs" $! #true with "[-]") as "HQvs".
              iLeft. iLeft. iSplitR. eauto. iExists [],[]. simpl. iFrame.
              iSplitR=>//. iSplitR "Hpg". iExists NONEV,None,None. iFrame.
              iRight. iSplitR. auto. iFrame; iFrame "#". iSplitL; eauto.
              iSplitL. iExists v'. iFrame; iFrame "#". done.
              iMod "HQvs". iModIntro. wp_seq.
              iExists ([p]). iFrame.
              
            - (* Absurd case *)
              by iDestruct (queue_dll_tail_some γp with "Hqlist1") as (e') "%"=>//.
          }

          { (* "prev" ≠ NONE *)
            subst. destruct p'=>//. destruct l1=>//. (* simpl. *)
            { (* Absurd case *)
              iDestruct "Hqlist1" as %Habs. inversion Habs. inversion H0. }

            (* next (unSOME "prev") <- "next")  *)
            iDestruct (queue_dll_do_not_contain  with "[$Hpg $Hqlist1]") as %Hpnotinl1.
            iDestruct (queue_dll_tail_elem_of with "Hqlist1") as %Hpinl1=>//. 
            iDestruct (queue_dll_last with "Hqlist1") as
                (l' prev v0 p0' n0') "(% & % & Hprevg & Hqlist1')"; auto. 
            inversion H0; subst.
            iDestruct (big_sepL_app _ (p1::l1) ([p]) with "Hprocinvs")
              as "(Hpinvsl1' & Hpinvsl2'')".
            rewrite H.
            iDestruct (big_sepL_app _ (l') ([prev]) with "Hpinvsl1")
              as "(Hpinvsl1'' & Hprev)". simpl.
            iDestruct "Hprev" as "(Hprev' & _)".
            
            wp_apply unSOME_spec.
            wp_apply next_spec.
            iApply (wp_proc_store_next N γp γq _ _ _ _ _ None). iFrame; iFrame "#".
            iIntros "Hprevg". wp_seq.

            (* (prev "p") <- NONE  *)
            wp_apply prev_spec.
            wp_apply (wp_proc_store_prev N γp γq _ _ _ _ _ None). iFrame; iFrame "#".
            iIntros "Hpg". wp_seq.

            (* (next "p") <- NONE  *)
            wp_apply next_spec.
            wp_apply (wp_proc_store_next N γp γq _ _ _ _ _ None). iFrame; iFrame "#".
            iIntros "Hpg". wp_seq.

            (* (qref "p") <- NONE  *)
            wp_apply qref_spec.
            wp_apply (wp_proc_store_queue_none). iFrame; iFrame "#". 
            iIntros "Hret". iDestruct "Hret" as (γ'') "(Hpg & Hplink & Hinq)".
            wp_seq.

            (* "head" <- !"h" *)
            wp_load. wp_bind (_ <- _)%E.
            iDestruct (queue_dll_singleton with "Hprevg") as "Hprevsingle".
            iDestruct (queue_dll_combine with "[Hqlist1' Hprevsingle]") as "Hqlist1".
            iSplitL "Hqlist1'". iFrame. iFrame.
            iApply (wp_mask_mono E); first done.
            iMod ("Hvs" with "HP") as (l0) "[Hq [_ HQvs]]".
            iDestruct "Hq" as (hv0 hvOP tvOP)
                                "(>Hhead' & Hsent & >Hlink & >Hlist' & Hrest)".
            iDestruct (link_eq with "[$Hlink $Hplink]") as %Hγeq; subst.
            iDestruct (list_combine with "[$Hlist $Hlist']") as %Hleq; subst.
            iDestruct (mapsto_agree (qhead q) with "Hhead Hhead'") as %foo; subst.
            iCombine "Hhead" "Hhead'" as "Hhead".
            wp_store.
            rewrite -H.
            iDestruct (list_update γ (p1 :: l1 ++ [p]) (p1 :: l1)
                       with "[$Hlist $Hlist']") as ">(Hlist & Hlist')".
            iDestruct "Hhead" as "[Hhead Hhead']".
            iDestruct (inQueue_remove p (p1::l1) [] with "[$Hqga $Hinq]") as ">Hqga".
            { rewrite disjoint_singleton_l app_nil_r.
              by apply not_elem_of_of_list'. }
            iDestruct ("HQvs" $! #true with "[-]") as "HQvs".
            iLeft. iLeft. iSplitR. eauto. iExists (p1::l1),[]. 
            iSplitR=>//. iSplitR "Hpg". 
            iExists (option_procAddrT_to_val (Some hv'')),(Some hv''),(Some prev). 
            rewrite app_nil_r. iFrame. iRight.
            iSplitR=>//. iFrame; iFrame "#". iSplitL.
            iExists v'. iFrame; iFrame "#". done.
            iMod "HQvs". iModIntro. wp_seq.
            iExists (p1 :: l1 ++ [p]). iFrame. }

        * (* "next" ≠ NONE *)
          subst. destruct n'=>//. 
          iDestruct (queue_dll_do_not_contain with "[$Hpg $Hqlist1]") as %Hpnotinl1.
          iDestruct (queue_dll_do_not_contain with "[$Hpg $Hqlist2]") as %Hpnotinl2.
          iDestruct (queue_dll_not_nil_head with "Hqlist2") as %Hl2notnil=>//.
          destruct l2=>//. 
          rewrite queue_dll_unfold.
          iDestruct "Hqlist2" as (v1 n1) "(Hpeq & Hn1g & Hqlist2')".
          iDestruct "Hpeq" as %Hpeq.
          inversion Hpeq=>//. subst.
          iDestruct (big_sepL_cons with "Hpinvsl2'") as "(Hn1inv & Hpinvsl2'')".

          (* prev (unSOME "next") <- "prev")  *)
          wp_apply unSOME_spec.
          wp_apply prev_spec.
          wp_apply (wp_proc_store_prev N γp γq _ _ _ _ _ p'). iFrame; iFrame "#".
          iIntros "Hn1g". wp_seq.

          iDestruct (queue_dll_singleton with "Hn1g") as "Hn1single".
          iDestruct (queue_dll_combine with "[Hqlist2' Hn1single]") as "Hqlist2".
          iSplitL "Hn1single". iFrame. iFrame.
          
          (* (if: "prev" = NONE then *)
          wp_op=>[Hprevnone|Hprevnotnone]; wp_if.
          
          { (* "prev" = NONE *)
            subst. destruct p'=>//. destruct l1=>//.
            
            - (* "h" <- "next" *)
              wp_store.

              (* (prev "p") <- NONE  *)
              wp_apply prev_spec.
              wp_apply (wp_proc_store_prev N γp γq _ _ _ _ _ None). iFrame; iFrame "#".
              iIntros "Hpg". wp_seq.

              (* (next "p") <- NONE  *)
              wp_apply next_spec.
              wp_apply (wp_proc_store_next N γp γq _ _ _ _ _ None). iFrame; iFrame "#".
              iIntros "Hpg". wp_seq.

              (* (qref "p") <- NONE  *)
              wp_apply qref_spec.
              wp_apply (wp_proc_store_queue_none). iFrame; iFrame "#". 
              iIntros "Hret". iDestruct "Hret" as (γ'') "(Hpg & Hplink & Hinq)".
              wp_seq.

              (* "head" <- !"h" *)
              wp_load.
              wp_bind (_ <- _)%E.
              iApply (wp_mask_mono E); first done.
              iMod ("Hvs" with "HP") as (l') "[Hq [_ HQvs]]".
              iDestruct "Hq" as (hv0 hvOP tvOP)
                                  "(>Hhead' & Hsent & >Hlink & >Hlist' & Hrest)".
              iDestruct (link_eq with "[$Hlink $Hplink]") as %Hγeq; subst.
              iDestruct (list_combine with "[$Hlist $Hlist']") as %Hleq; subst.
              iDestruct (mapsto_agree (qhead q) with "Hhead Hhead'") as %foo; subst.
              iCombine "Hhead" "Hhead'" as "Hhead".
              wp_store.
              iDestruct (list_update γ (p :: p1 :: l2) (p1 :: l2)
                         with "[$Hlist $Hlist']") as ">(Hlist & Hlist')".
              iDestruct "Hhead" as "[Hhead Hhead']".
              iDestruct (inQueue_remove p [] (p1 :: l2) with "[$Hqga $Hinq]")
                as ">Hqga".
              { rewrite disjoint_singleton_l. by apply not_elem_of_of_list'. }
              iDestruct ("HQvs" $! #true with "[-]") as "HQvs".
              iLeft. iLeft. iSplitR. eauto. iExists [],(p1 :: l2). iFrame.
              iSplitR=>//. iSplitR "Hpg". 
              iExists (option_procAddrT_to_val (Some p1)),(Some p1),tv. iFrame.
              iRight. iSplitR. auto. iFrame; iFrame "#". eauto.
              iSplitL. iExists v'. iFrame; iFrame "#". done.
              iMod "HQvs". iModIntro. wp_seq.
              iExists (p :: p1 :: l2). iFrame.
              
            - (* Absurd case *)
              iDestruct (queue_dll_tail_some with "Hqlist1") as (e') "%"=>//.
          }

          { (* "prev" ≠ NONE *)
            subst. destruct p'=>//. destruct l1=>//.
            { (* Absurd case *)
              iDestruct "Hqlist1" as %Habs. inversion Habs. inversion H0. }

            (* next (unSOME "prev") <- "next")  *)
            iDestruct (queue_dll_tail_elem_of with "Hqlist1") as %Hpinl1=>//. 
            iDestruct (queue_dll_last with "Hqlist1") as
                (l' prev v0 p0' n0') "(% & % & Hprevg & Hqlist1')"; auto.
            inversion H0; subst.
            iDestruct (big_sepL_elem_of with "Hpinvsl1") as "Hprevinv"=>//. 
            wp_apply unSOME_spec.
            wp_apply next_spec.
            wp_apply (wp_proc_store_next N γp γq _ _ _ _ _ (Some p1)). 
            iFrame; iFrame "#". iIntros "Hprevg". wp_seq.

            (* (prev "p") <- NONE  *)
            wp_apply prev_spec.
            wp_apply (wp_proc_store_prev N γp γq _ _ _ _ _ None). iFrame; iFrame "#".
            iIntros "Hpg". wp_seq.

            (* (next "p") <- NONE  *)
            wp_apply next_spec.
            wp_apply (wp_proc_store_next N γp γq _ _ _ _ _ None). iFrame; iFrame "#".
            iIntros "Hpg". wp_seq.

            (* (qref "p") <- NONE  *)
            wp_apply qref_spec.
            wp_apply (wp_proc_store_queue_none). iFrame; iFrame "#". 
            iIntros "Hret". iDestruct "Hret" as (γ'') "(Hpg & Hplink & Hinq)".
            wp_seq.

            (* "head" <- !"h" *)
            wp_load. wp_bind (_ <- _)%E.
            iDestruct (queue_dll_singleton with "Hprevg") as "Hprevsingle".
            iDestruct (queue_dll_combine with "[Hqlist1' Hprevsingle]") as "Hqlist1".
            iSplitL "Hqlist1'". iFrame. iFrame.
            iDestruct (queue_dll_combine with "[Hqlist1 Hqlist2]") as "Hqlist".
            iSplitL "Hqlist1". iFrame. iFrame.
            iApply (wp_mask_mono E); first done.
            iMod ("Hvs" with "HP") as (l0) "[Hq [_ HQvs]]".
            iDestruct "Hq" as (hv0 hvOP tvOP)
                                "(>Hhead' & Hsent & >Hlink & >Hlist' & Hrest)".
            iDestruct (link_eq with "[$Hlink $Hplink]") as %Hγeq; subst.
            iDestruct (list_combine with "[$Hlist $Hlist']") as %Hleq; subst.
            iDestruct (mapsto_agree (qhead q) with "Hhead Hhead'") as %foo; subst.
            iCombine "Hhead" "Hhead'" as "Hhead".
            wp_store.
            iDestruct (list_update γ (p2 :: l1 ++ p :: p1 :: l2) (p2 :: l1 ++ p1 :: l2)
                       with "[$Hlist $Hlist']") as ">(Hlist & Hlist')".
            iDestruct "Hhead" as "[Hhead Hhead']".
            iDestruct (inQueue_remove p (p2 :: l1) (p1 :: l2) with "[$Hqga $Hinq]")
              as ">Hqga".
            { rewrite disjoint_singleton_l.
              by apply not_elem_of_of_list', not_elem_of_app. }
            iDestruct ("HQvs" $! #true with "[-]") as "HQvs".
            iLeft. iLeft. iSplitR. eauto. iExists (p2::l1),(p1::l2). 
            iSplitR=>//. iSplitR "Hpg". 
            iExists (option_procAddrT_to_val (Some hv'')),(Some hv''),tv. 
            iFrame. iRight. iSplitR=>//. iFrame; iFrame "#". rewrite -H. iFrame.
            iSplitL. iExists v'. iFrame; iFrame "#". eauto.
            iMod "HQvs". iModIntro. wp_seq.
            iExists (p2 :: l1 ++ p :: p1 :: l2). iFrame. }

        * (* Absurd case *)
          by destruct Hneq'. 

        * (* p ∉ l *)
          wp_apply qref_spec.
          wp_bind (!_)%E.
          iApply (wp_mask_mono E); first done.
          iMod ("Hvs" with "HP") as (l') "[Hq [HPvs _]]".
          iDestruct "Hq" as (hv0 hvOP tvOP) "(Hqhead & Hlist' & >Hlink & Hrest)".
          iInv (n_inv_proc N p) as (pvv' pqv' ppv' pnv')
             ">(Hpvl & Hpql & Hppl & Hpnl & Hinq)" "HcloseProcInv".
          destruct (decide (pqv' = (Some q))).
          { rewrite e.
            iDestruct "Hinq" as (γ'') "(Hplink & Hinq)".
            iDestruct (link_eq with "[$Hlink $Hplink]") as %Hγeq; subst.
            iDestruct (inQueue_combine with "[$Hqga $Hinq]") as %Hpinl.
            contradiction. }
          wp_load. iMod ("HcloseProcInv" with "[Hpvl Hpql Hppl Hpnl Hinq]") as "_".
          { iNext. iExists pvv',pqv',ppv',pnv'; by iFrame. }
          iModIntro.
          iMod ("HPvs" with "[Hqhead Hlist' Hlink Hrest]") as "HP".
          iExists hv0,hvOP,tvOP. iFrame. iModIntro.
          (* iApply fupd_intro_mask. set_solver. *)
          wp_op=>[Habs|Hnotinl]; wp_if.

          { destruct n.
            destruct pqv'=>//.
            assert (SOMEV (#(qhead q), (#(qtail q), #(qsent q)))
                    = option_queueAddrT_to_val (Some q)).
            { done. } rewrite H in Habs.
            by apply option_queueAddrT_to_val_eq in Habs. }
            
          (* "head" <- !"h" ;; #false. *)
          wp_load.
          wp_bind (_ <- _)%E.
          iApply (wp_mask_mono E); first done.
          iMod ("Hvs" with "HP") as (l0) "[Hq [_ HQvs]]".
          iDestruct "Hq" as (hv0' hvOP' tvOP')
                              "(>Hhead' & Hsent & Hlink & >Hlist' & Hrest)".
          iDestruct (list_combine with "[$Hlist $Hlist']") as %Hleq; subst.
          iDestruct (mapsto_agree (qhead q) with "Hhead Hhead'") as %foo; subst.
          iCombine "Hhead" "Hhead'" as "Hhead".
          wp_store.
          iDestruct "Hhead" as "[Hhead Hhead']".
          iDestruct ("HQvs" $! #false with "[-]") as "HQvs".
          iLeft. iRight. iSplitR=>//. iFrame. iSplitL.
          iExists (option_procAddrT_to_val (Some hv'')),(Some hv''),tv. 
          iFrame. iRight. iFrame; iFrame "#". done. done.
          iMod "HQvs". iModIntro. wp_seq.
          iExists l0. iFrame. 
  Qed.

End dequeue.

 