From iris.algebra Require Import gmap frac deprecated auth gset.
From iris.base_logic Require Import big_op wsat invariants fractional.
From iris.proofmode Require Import tactics.
From iris.heap_lang Require Import proofmode notation.
From stdpp Require Import mapset.
From fletch Require Import wp_helpers helpers definitions monoids invariants
     atomic program.

Import dec_agree.

Section enqueue.

  Context {Σ : gFunctors} `{!heapG Σ, !fletchG Σ}
          (N : namespace) (γp γq γl : gname).
    
  Local Notation iProp := (iProp Σ).

  Definition enq_α q γ γ' := (fun l => ▷ queue N γp γq q γ γ' l)%I.
  Definition enq_β q γ γ' p := (fun l ret => queue N γp γq q γ γ' (l++[p]) ∗
                                          ⌜ret = #(is_nil l)⌝)%I.

  Definition enqueue_atomic_spec (q : queueAddrT) (γ γ' : gname)
             (p : procAddrT) P Q E :=
    atomic_shift (enq_α q γ γ') (enq_β q γ γ' p) ∅ E P Q.
  
  Lemma obtainLockEnq_spec (q : queueAddrT) (γ γ' : gname) (h : loc) p P Q E φ :
    enqueue_atomic_spec q γ γ' p P Q E ∗ 
    (∃ hv, h ↦ hv) ∗ P ∗
    (∀ l hv' tv, P ∗ h ↦ (option_procAddrT_to_val hv') ∗ 
        (queue_lock N γp γq q γ γ' (SOMEV #(qsent q)) hv' tv l) -∗ φ #())
     ⊢ WP obtainLockEnq #(qhead q) #(qsent q) #h {{ φ }}.
  Proof.
    iIntros "(#Hvs & Hh & HP & Hret)".
    iLöb as "IH".
    iDestruct "Hh" as (hv) "Hh".
    destruct q; simpl.
    wp_rec. wp_let. wp_let. wp_load. wp_let.
    wp_op=>[->|?]. wp_if.
    - (* h points to sentinel *)
      wp_if. wp_bind (!_)%E.
      iApply (wp_mask_mono E); first done.
      iMod ("Hvs" with "HP") as (l) "[Hq [HPvs _]]".
      iDestruct "Hq" as (hv0 hvOP tvOP) "(Hqhead & Hrest)". simpl.
      wp_load.
      iMod ("HPvs" with "[Hqhead Hrest]") as "HP". 
      iExists hv0,hvOP,tvOP. iFrame. iModIntro.
      wp_store.
      iApply ("IH" with "[Hh] HP Hret"). eauto.
    - (* h do not point to sentinel, try CAS *)
      wp_if. wp_bind (CAS _ _ _)%E.
      iApply (wp_mask_mono E); first done.
      iMod ("Hvs" with "HP") as (l) "[Hq [HPvs _]]".
      iDestruct "Hq" as (hv0 hvOP tvOP) "(>Hqhead & >Hlist & Hrest)". simpl.
      destruct (decide (hv = hv0)) as [->|Hneq].
      + (* CAS will succeed, since hv0 is not sentinel *)
        iDestruct "Hrest" as
            "(>Hlink & >Hlist' & [>Hsent| [>HvOp [>Hqhead' Hlock]]])"=>//.
        iDestruct "Hsent" as %Habsurd=>//. simpl.
        iDestruct "HvOp" as %HvOp.
        iCombine "Hqhead" "Hqhead'" as "Hqhead".
        wp_cas_suc.
        iDestruct "Hqhead" as "[Hqhead Hqhead']".
        iMod ("HPvs" with "[Hqhead Hlist Hlist' Hlink]") as "HP".
        iExists (SOMEV #qsent),hvOP,tvOP. iFrame. iLeft. auto.
        iModIntro. wp_op. wp_if.
        iApply ("Hret" $! l hvOP tvOP with "[-]"); subst. iFrame.
      + (* CAS will not succeed *)
        wp_cas_fail.
        iMod ("HPvs" with "[Hqhead Hlist Hrest]") as "HP".
        iExists hv0,hvOP,tvOP. iFrame.
        iModIntro. wp_op.
        wp_if. wp_bind (!_)%E.
        iApply (wp_mask_mono E); first done.
        iMod ("Hvs" with "HP") as (l') "[Hq [HPvs _]]".
        iDestruct "Hq" as (hv0' hvOP' tvOP') "(Hqhead & Hrest)". simpl.
        wp_load.
        iMod ("HPvs" with "[Hqhead Hrest]") as "HP". 
        iExists hv0',hvOP',tvOP'. iFrame. iModIntro.
        wp_store.
        iApply ("IH" with "[Hh] HP Hret"). eauto.
  Qed.
  
  Lemma enqueue_spec :
    ∀ q p γ γ' E,
      procs_inv N γp ∗
      qproc N γp γq p ⊢
      atomic_wp
      (enq_α q γ γ')
      (enq_β q γ γ' p)
      ∅ E
      (enqueue #(qhead q) #(qtail q) #(qsent q) (procAddrT_to_val p))%E.
  Proof.
    iIntros (q p γ γ' E) "(#Hprocsinv & Hqproc)".
    iIntros (P Q) "#Hvs HP".
    iDestruct "Hqproc" as (v) "(Hpg & #Hprocinv)".
    wp_lam. wp_lam. wp_let. wp_let.

    (* let: "h" := ref !"head" in *)
    wp_bind (! _)%E.
    iApply (wp_mask_mono E); first done.
    iMod ("Hvs" with "HP") as (l) "[Hq [HPvs _]]".
    iDestruct "Hq" as (hv0 hvOP tvOP) "(Hqhead & Hqsent & Hlink & Hrest)".
    iDestruct (link_split with "Hlink") as "(Hlink & Hlink')".
    wp_load.
    iMod ("HPvs" with "[Hqhead Hqsent Hlink Hrest]") as "HP". 
    iExists hv0,hvOP,tvOP. iFrame. iModIntro.
    wp_alloc h as "Hh". wp_let. clear hvOP tvOP l.

    (* let: "obtLock" := tryObtainLock "head" "s" "h" in *)
    wp_apply (obtainLockEnq_spec q γ γ' h p P Q with "[-]").
    iFrame; iFrame "#". iSplitL "Hh". by iExists hv0. 
    iIntros (l hv' tv)
            "(HP & Hh & (Hhead & Hlist & Hqga & Htail & #Hprocinvs & Hqlist))".
    wp_seq.

    (* Allocate new token *)
    iDestruct (queue_dll_do_not_contain with "[Hpg Hqlist]") as %Hpnotinl; iFrame.
    iDestruct (inQueue_add _ p with "Hqga") as "Hvstok". done.
    (* { by apply not_elem_of_of_list' in Hpnotinl. } *)
    iMod "Hvstok". iDestruct "Hvstok" as "(Hqga & Hpgf)".

    (* qref "p" <- SOME ("head",("tail","s")) *)
    wp_apply qref_spec.
    have->:SOME (#(qhead q), (#(qtail q), #(qsent q))) =
      option_queueAddrT_to_val (Some q) by (by destruct q as [? []]).
    wp_apply wp_proc_store_queue_some. iFrame; iFrame "#".
    iSplitL "Hlink' Hpgf". iExists γ'. iFrame.
    iIntros "Hpg". wp_seq.

    (* match: !"h" with *)
    wp_bind (!_)%E. wp_load.
    destruct hv' as [ph|]; rewrite /option_procAddrT_to_val; simpl; wp_match.

    - (* CASE SOME "h'" => *)
      iDestruct (queue_dll_not_nil_head with "Hqlist") as %Hlnotnil. auto.
      iDestruct (queue_dll_tail_some with "Hqlist") as (e') "%"; auto. subst.
      iDestruct (queue_dll_tail_elem_of with "Hqlist") as %He'inl. auto.
      iDestruct (queue_dll_last with "Hqlist") as
          (l' e'' v0 p0 n0) "(Hl & He'e'' & Hng & Hqlistl')"; auto.
      iDestruct "Hl" as %Hl.
      iDestruct "He'e''" as %He'e''.
      
      (* prev "p" <- !"tail" *)
      wp_apply prev_spec. wp_load.
      wp_bind (_ <- _)%E.
      rewrite !option_procAddrT_to_val_some.
      iApply (wp_proc_store_prev). iFrame; iFrame "#".
      iIntros "Hpg". wp_seq.

      (* next (unSOME !"tail") <- SOME "p" *)
      wp_load. wp_apply unSOME_spec. wp_apply next_spec.
      iDestruct (big_sepL_elem_of _ l e' with "Hprocinvs") as "Hpinv". done.
      inversion He'e''. subst.
      wp_bind (_ <- _)%E.
      rewrite !option_procAddrT_to_val_some.
      iApply wp_proc_store_next. iFrame "#". iFrame. iIntros "Hng". wp_seq.
        
      (* "tail" <- SOME "p" *)
      wp_store.

      (* "head" <- SOME "h'" *)
      wp_bind (_ <- _)%E.

      (* Release the lock with l ++ [p] as the list *)
      iDestruct (queue_dll_singleton with "Hng") as "Hqn".
      iDestruct (queue_dll_singleton with "Hpg") as "Hqp".
      iDestruct (queue_dll_combine with "[Hqlistl' Hqn]") as "Hqlist".
      iSplitL "Hqlistl'"; by iFrame.
      iDestruct (queue_dll_combine with "[Hqlist Hqp]") as "Hqlist".
      iSplitL "Hqlist". iFrame. iFrame.

      iApply (wp_mask_mono E); first done.
      iMod ("Hvs" with "HP") as (l) "[Hq [_ HQvs]]".
      iDestruct "Hq" as (hv0' hvOP' tvOP')
                          "(>Hhead' & Hsent & Hlink & Hlist' & Hrest)".

      iDestruct (mapsto_agree (qhead q) with "Hhead Hhead'") as %foo. subst.
      iCombine "Hhead" "Hhead'" as "Hhead".
      wp_store.
      iDestruct (list_combine with "[$Hlist $Hlist']") as %Heq=>//; subst.
      iDestruct (list_update γ (l' ++ [e'']) ((l' ++ [e'']) ++ [p])
                 with "[$Hlist $Hlist']") as ">(Hlist & Hlist')".
      iDestruct "Hhead" as "[Hhead Hhead']".
      
      iMod ("HQvs"  $! #false with "[-]") as "HQ".
      iSplitL.
      iExists (option_procAddrT_to_val (Some ph)),(Some ph),(Some p).
      iFrame. iRight. iSplitR=>//. 
      iCombine "Hprocinvs" "Hprocinv" as "Hprocinvs'".
      rewrite (proc_singleton N γp γq p) -big_sepL_app.
      iFrame; iFrame "#". by rewrite is_nil_app.
      iModIntro. wp_seq. by iExists (l' ++ [e'']).

    -  (* CASE NONE => *)
      iDestruct (queue_dll_none with "Hqlist") as %Hlnil; subst.
      (* "tail" <- SOME "p" *)
      wp_store.
      (* "head" <- SOME "p" *)
      wp_bind (_ <- _)%E.
      (* Release the lock with [p] as the list *)        
      iDestruct (queue_dll_singleton with "Hpg") as "Hqlist'".

      iApply (wp_mask_mono E); first done.
      iMod ("Hvs" with "HP") as (l) "[Hq [_ HQvs]]".
      iDestruct "Hq" as (hv0' hvOP' tvOP')
                          "(>Hhead' & Hsent & Hlink & Hlist' & Hrest)".
      iDestruct (mapsto_agree (qhead q) with "Hhead Hhead'") as %foo; subst.
      iCombine "Hhead" "Hhead'" as "Hhead".
      wp_store.
      iDestruct (list_combine with "[$Hlist $Hlist']") as %Heq=>//; subst.
      iDestruct (list_update γ ([]) ([p])
                 with "[$Hlist $Hlist']") as ">(Hlist & Hlist')".
      iDestruct "Hhead" as "[Hhead Hhead']".
      rewrite proc_singleton.      
      iMod ("HQvs" $! #true with "[-]") as "HQ". 
      iSplitL. 
      iExists (option_procAddrT_to_val (Some p)),(Some p),(Some p).
      iFrame. iRight. iSplitR=>//. iFrame; iFrame "#". by unfold is_nil. 
      iModIntro. wp_seq. by iExists ([]).
  Qed.

End enqueue.
