From iris.algebra Require Import gmap frac deprecated auth gset.
From iris.base_logic Require Import lib.gen_heap big_op wsat invariants.
From iris.proofmode Require Import tactics.
From iris.heap_lang Require Import proofmode notation.
From stdpp Require Import mapset.
From fletch Require Import definitions monoids invariants.

Import dec_agree.

Section helpers.

  Context {Σ : gFunctors} `{!heapG Σ, !fletchG Σ}
          (N : namespace) (γp γq γl : gname).
  Context `{heapN ⊥ N}.
    
  Local Notation iProp := (iProp Σ).

  Lemma proc_own_update p v q q' pv pv' nv nv' E :
    nclose (n_inv_procs N) ⊆ E →
    procs_inv N γp ∗ own γp (Proc p 1 {| pvalv := v;
                                         pqueuev := q;
                                         pprevv := pv;
                                         pnextv := nv; |})
    ⊢ |={E}=> own γp (Proc p 1 {| pvalv := v;
                                  pqueuev := q';
                                  pprevv := pv';
                                  pnextv := nv'; |}).
  Proof.
    iIntros (Hns) "(#Hprocs & Hpg)".
    iInv (n_inv_procs N) as (h) ">(Hpa & Hmap)" "Hclose".
    iDestruct (proc_own_contains with "[$Hpa $Hpg]") as %[π Hsome].
    rewrite (big_sepM_delete _ h p _); last done.  
    iDestruct "Hmap" as "(Hpv & Hmap)". iDestruct "Hpv" as (v0') "(% & Hpv)".
    iDestruct (proc_own_update_ghost γp h p {| pvalv := v;
                                         pqueuev := q;
                                         pprevv := pv;
                                         pnextv := nv; |}
                                     {| pvalv := v;
                                        pqueuev := q';
                                        pprevv := pv';
                                        pnextv := nv'; |} 
               with "[$Hpa $Hpg]") as ">(Hpa & Hpg)". iFrame.
    iAssert ([∗ map] k↦y ∈ (<[p:=(1%Qp, DecAgree {| pvalv := v;
                                        pqueuev := q';
                                        pprevv := pv';
                                        pnextv := nv'; |})]>  h),
             ∃ v0 : procValT, ⌜y = (1%Qp, DecAgree v0)⌝ ∗ pvall k ↦{½} pvalv v0)%I
    with "[Hmap Hpv]" as "Hmap".
    { rewrite -insert_delete.
      rewrite big_sepM_insert.
      - iFrame. iExists {| pvalv := v; pqueuev := q'; pprevv := pv'; pnextv := nv' |}.
        inversion H0; subst; simpl.
          by iFrame. 
      - apply lookup_delete. }
    iMod ("Hclose" with "[Hmap Hpa]") as "_".
    iNext. iExists (<[p:=(1%Qp, DecAgree _)]> h). iFrame.
    by iModIntro. 
  Qed.

  Lemma proc_singleton p :
    proc_inv N γp γq p ⊢ [∗ list] x ∈ [p], proc_inv N γp γq x.
  Proof.
    iIntros. by rewrite big_sepL_singleton.
  Qed.

(* IS_QUEUEDLIST
===================================================*)

  Lemma queue_dll_unfold q x l i e p n :
    queue_dll γp q (x :: l) i e p n =
    (∃ (v : val) (n' : option procAddrT),
        ⌜i = Some x⌝ ∗ own γp (Proc x ¾ {| pvalv := v;
                                           pqueuev := Some q;
                                           pprevv := p;
                                           pnextv := n' |}) ∗
       queue_dll γp q l n' e i n)%I.
  Proof. reflexivity. Qed.

  Lemma queue_dll_tail_elem_of q l x n :
    l ≠ [] → ∀ i p, queue_dll γp q l i (Some x) p n
    ⊢ ⌜(x ∈ l)⌝.
  Proof.
    induction l; intro Hnotempty; intros i p.
    - by destruct Hnotempty.
    - destruct l; iIntros "H".
      + iDestruct "H" as (v0 n0) "(% & Hpg & %)"; destruct H1; subst. inversion H2.
        iPureIntro. apply elem_of_list_here.
      + destruct (decide (x = a)); subst.
        * iPureIntro. apply elem_of_list_here.
        * rewrite queue_dll_unfold.
          iDestruct "H" as (v0' n0') "(% & Hpg & Hqlist)".
          iDestruct (IHl with "Hqlist") as %Hfurther. auto.
          iPureIntro. by apply elem_of_list_further.
  Qed.

  Lemma queue_dll_singleton x v q p n :
    own γp (Proc x ¾ {| pvalv := v; pqueuev := (Some q); pprevv := p; pnextv := n |})

    ⊢ queue_dll γp q [x] (Some x) (Some x) p n.
  Proof.
    iIntros "H". iExists v,n. iFrame. auto. 
  Qed.

  Lemma queue_dll_split x l q  :
    x ∈ l → ∀ h t p n, queue_dll γp q l h t p n
    ⊢ ∃ v p' n' l1 l2, ⌜l = l1 ++ x :: l2⌝ ∗
                      own γp (Proc x ¾ (ProcValT v (Some q) p' n')) ∗ 
                     queue_dll γp q l1 h p' p (Some x) ∗
                     queue_dll γp q l2 n' t (Some x) n.
   Proof.
     intro Hinl; apply elem_of_list_split in Hinl; destruct Hinl as [l1 [l2]]; subst.
     induction l1 as [| h0 l1]; intros; simpl; iIntros "Hqlist".
     -  iDestruct "Hqlist" as (v' n') "(% & Hpg & Hqlistl2)".
        iExists v',p,n',[],l2; subst; iFrame; auto.
     -  iDestruct "Hqlist" as (v' n') "(% & Hpg & Hqrest)".
        iDestruct (IHl1 with "Hqrest") as (v'' p'' n'' l1' l2')
                                            "(% & Hxg & Hqlist1 & Hqlist2)". 
        iExists v'',p'',n'',(h0::l1'),l2'. iFrame. iSplitR.
        rewrite H1 ; auto.
        iExists v',n'. by iFrame.
   Qed.

  Lemma queue_dll_combine q l1 l2 h' e' e n :
    ∀ h p, queue_dll γp q l1 h e' p h' ∗
    queue_dll γp q l2 h' e e' n
    ⊢ queue_dll γp q (l1 ++ l2) h e p n.
  Proof.
    induction l1; intros; iIntros "(Hl1 & Hl2)"; simpl.
    - by iDestruct "Hl1" as %[h1 h2]; subst. 
    - iDestruct "Hl1" as (v0 n0) "(% & Hpg & Hql1)"; subst.
      iDestruct (IHl1 with "[Hql1 Hl2]") as "Hl"; iFrame. 
      iExists v0,n0. iFrame=>//. 
  Qed.      
  
  Lemma queue_dll_nil_head x v q l e p n :
    own γp (Proc x ¾ v) ∗ queue_dll γp q l (Some x) e p n
    ⊢ ⌜l = []⌝.
  Proof.
    iIntros "(Hpg & Hqlist)".
    destruct l. 
    - auto.
    - iDestruct "Hqlist" as (v' n') "(% & Hpg' & Hqrest)".
      inversion H0; subst.
      iDestruct (proc_too_many_fractions γp p0 with "[Hpg Hpg']") as "[]". iFrame. 
  Qed.       

  Lemma queue_dll_nil_tail x v q l i p n :
    own γp (Proc x ¾ v) ∗ queue_dll γp q l i (Some x) p n
        ⊢ ⌜l = []⌝. 
  Proof.
    induction l; iIntros "(Hpg & Hqlist)".
    - auto.
    - iDestruct (queue_dll_tail_elem_of with "Hqlist") as %Hxinl; auto.
      iDestruct (queue_dll_split with "Hqlist") as
          (v0 p0 n0 l1 l2) "(% & Hpg' & Hrest)"=>//.
      iDestruct (proc_too_many_fractions γp x with "[Hpg Hpg']") as "[]". iFrame.
  Qed.

  Lemma queue_dll_do_not_contain x v q l :
    ∀ i e p n,
      own γp (Proc x ¾ v) ∗ queue_dll γp q l i e p n
          ⊢ ⌜(x ∉ l)⌝.
  Proof.
    induction l; intros; iIntros "(Hpg & Hqlist)".
    - iPureIntro. intro. inversion H0.
    - rewrite queue_dll_unfold.
      iDestruct "Hqlist" as (v0 n0) "(% & Hpg' & Hqlist)". 
      destruct (decide (x = a)). subst.
      + iDestruct (proc_too_many_fractions with "[Hpg Hpg']") as %Hfalse. 
        iFrame. inversion Hfalse.
      + iDestruct (IHl with "[Hpg Hqlist]") as %Hnotin. iFrame. 
        iPureIntro. apply not_elem_of_cons. auto.
  Qed.
   
  Lemma queue_dll_not_nil_tail q l i e p n :
    e ≠ (Some p) → queue_dll γp q l i e (Some p) n
                             ⊢ ⌜(l ≠ [])⌝.
  Proof.
    intros. destruct l; iIntros "Hqlist".
    - iDestruct "Hqlist" as %[H1 H2]. contradiction.
    - iPureIntro. auto.
  Qed.

  Lemma queue_dll_tail_some q l e n :
    ∀ h p, l ≠ [] → queue_dll γp q l h e p n ⊢ ∃ e', ⌜e = Some e'⌝.
  Proof.
    induction l.
    - destruct 1. done.
    - destruct l; intros; iIntros "H".
      + iDestruct "H" as (v0 n0) "(% & Hpg & %)".
        destruct H2. subst. iExists a=>//. 
      + rewrite queue_dll_unfold.
        iDestruct "H" as (v0 n0) "(% & Hpg & Hqlist)".
        iApply IHl. done. iFrame. 
  Qed.    

  Lemma queue_dll_not_nil_head q l h e p n :
     (Some h) ≠ n →
     queue_dll γp q l (Some h) e p n
     ⊢ ⌜l ≠ []⌝.
   Proof.
     intros. destruct l; iIntros "Hqlist".
     - iDestruct "Hqlist" as %[H1 H2]. contradiction.
     - iPureIntro. auto.
   Qed.

   Lemma queue_dll_last q l e n :
     l ≠ [] →
     ∀ h p, queue_dll γp q l h e p n
     ⊢ ∃ l' e' v' p' n', ⌜l = l' ++ [e']⌝ ∗ ⌜e = Some e'⌝ ∗
                        own γp (Proc e' ¾ (ProcValT v' (Some q) p' n')) ∗ 
                        queue_dll γp q l' h p' p (Some e').
   Proof.
     induction l; intros; iIntros "H".
     - done. 
     - destruct l. 
       + iDestruct "H" as (v0 n0) "(% & HX & %)". destruct H2. subst. 
         iExists [],a,v0,p,n. iFrame. iSplit=>//. 
       + rewrite queue_dll_unfold. 
         iDestruct "H" as (v0 n0) "(% & HX & Hrest)". subst. 
         iDestruct (IHl with "Hrest") as "HXX". auto.
         iDestruct "HXX" as (l' e' v' p' n') "(% & % & Heg & Hqlist)".
         iExists (a :: l'),e',v',p',n'. iSplit. rewrite H1=>//. iSplit=>//. 
         iFrame. iExists v0,n0. iFrame=>//. 
   Qed.          

   Lemma queue_dll_none q l t p n : 
     queue_dll γp q l None t p n ⊢ ⌜l = []⌝.
   Proof.
     iIntros "H"; destruct l.
     - done.
     - rewrite queue_dll_unfold.
       iDestruct "H" as (v0 n0) "(% & _)".
       inversion H0.
   Qed.

   Lemma singleton_app_cons (p : procAddrT) :
    ∀ l1 l2, [p] = l1 ++ p :: l2 → l1 ++ l2 = [].
   Proof.
     destruct l1; intros l2 Hl. simpl in Hl.
     - inversion Hl. done.
     - inversion Hl. destruct l1; inversion H2.
   Qed.

  Lemma heap_mapsto_invalid (l : loc) (q1 q2 : Qp) (v v' : val) :
    ¬ (Qp_plus q1 q2 ≤ 1%Qp)%Qc →
    l ↦{q1} v ∗ l ↦{q2} v' ⊢ False.
  Proof.
    iIntros (X) "(Hl1 & Hl2)".
    iPoseProof (mapsto_valid_2 l with "Hl1 Hl2") as "HX".
    by iDestruct "HX" as %Xvalid. 
  Qed.

End helpers.