From iris.algebra Require Import gmap frac deprecated auth gset.
From iris.base_logic Require Import big_op wsat invariants.
From iris.proofmode Require Import tactics.
From iris.heap_lang Require Import proofmode notation.
From stdpp Require Import mapset.
From fletch Require Import definitions monoids.

Import dec_agree.

Section invariants.

  Context {Σ : gFunctors} `{!heapG Σ, !fletchG Σ}
          (N : namespace).
  Context `{heapN ⊥ N}.
  
(* INVARIANTS AND PREDICATES 
===================================================*)

  Definition open_proc_mask (p : procAddrT) := (⊤ ∖ nclose (N .@ "proc" .@ p)).

  Definition procs γp := 
    (∃ h, own γp (ProcAuth h) ∗
          [∗ map] p ↦ v ∈ h, ∃ v0, ⌜v = (1%Qp, DecAgree v0)⌝ ∗
                                   (pvall p) ↦{½} (pvalv v0))%I.

  Definition procs_inv γp := (inv (n_inv_procs N) (procs γp))%I.

  Definition proc γp γq (a : procAddrT) (v : val) (qv : option queueAddrT)
                  (pv nv : option procAddrT) :=
    (own γp (Proc a ¼ {| pvalv := v;
                         pqueuev := qv;
                         pprevv := pv;
                         pnextv := nv |}) ∗
    (pqueuel a) ↦ option_queueAddrT_to_val qv ∗
    (pprevl a) ↦ option_procAddrT_to_val pv ∗ 
    (pnextl a) ↦ option_procAddrT_to_val nv ∗
    match qv with
    | None => True
    | Some q => ∃ γ, own γq (Link (qsent q) γ) ∗ own γ (InQueue a)
     end)%I.

  Global Instance proc_timeless : TimelessP (proc γp γq a v qv pv nv).
  Proof. intros. apply _. Qed.

  Definition proc_inv γp γq (x : procAddrT)  :=
    inv (n_inv_proc N x) (∃ v q p n, proc γp γq x v q p n)%I.

  Definition qproc γp γq x :=
    (∃ v, own γp (Proc x ¾ {| pvalv := v;
                              pqueuev := None;
                              pprevv := None;
                              pnextv := None |}) ∗ proc_inv γp γq x)%I.

  Fixpoint queue_dll γp (q : queueAddrT) (l : list procAddrT) (i e p n : option procAddrT)  :=
    match l with
      | nil => ⌜(i = n ∧ e = p)⌝
      | x :: l' => ∃ (v : val) (n' : option procAddrT),
                   ⌜i = Some x⌝ ∗ own γp (Proc x ¾ {| pvalv := v;
                              pqueuev := (Some q);
                              pprevv := p;
                              pnextv := n' |}) ∗
                       queue_dll γp q l' n' e i n 
    end%I.
  
  Definition queue_cont γp γq (q : queueAddrT) (h t : option procAddrT)
             (l : list procAddrT) :=
    (([∗ list] x ∈ l, proc_inv γp γq x) ∗ queue_dll γp q l h t None None)%I.

  Definition queue_lock γp γq (q : queueAddrT) γ γ' (hv : val) 
             (hv' tv : option procAddrT) (l : list procAddrT) :=
    ((qhead q) ↦{½} hv ∗ own γ (List l) ∗ own γ' (InQueueAuth l) ∗
     (qtail q) ↦ option_procAddrT_to_val tv ∗ queue_cont γp γq q hv' tv l)%I.

  Definition queue γp γq (q : queueAddrT) γ γ' l :=
    (∃ (hv : val) (hvOP tvOP : option procAddrT),
       (qhead q) ↦{½} hv ∗ (qsent q) ↦{½} #() ∗ 
       own γq (Link (qsent q) γ') ∗ own γ (List l) ∗
       (⌜hv = SOMEV #(qsent q)⌝ ∨ (⌜hv = (option_procAddrT_to_val hvOP)⌝ ∗
         queue_lock γp γq q γ γ' hv hvOP tvOP l)))%I.
  
  Definition queues γq := 
    (∃ qs, own γq (LinkAll qs) ∗ 
           [∗ map] l ↦ _ ∈ qs, ∃ γ, own γq (Link l γ) ∗ l ↦{½} #())%I.

  Definition queues_inv γq := (inv (n_inv_queues N) (queues γq))%I.

  Global Instance proc_inv_persistent : PersistentP (proc_inv γp γq x).
  Proof. apply _. Qed.

  Global Instance procs_inv_persistent : PersistentP (procs_inv γp).
  Proof. apply _. Qed.

  Global Instance queues_inv_persistent : PersistentP (queues_inv γq).
  Proof. apply _. Qed.

End invariants.