From iris.algebra Require Import gmap frac deprecated auth gset.
From iris.base_logic Require Import big_op wsat invariants.
From iris.proofmode Require Import tactics.
From iris.heap_lang Require Import proofmode notation.
From stdpp Require Import mapset.
From fletch Require Import wp_helpers helpers definitions monoids invariants
     atomic program.

Import dec_agree.

Section makers.

  Context {Σ : gFunctors} `{!heapG Σ, !fletchG Σ}
          (N : namespace) (γp γq : gname).
  
  Local Notation iProp := (iProp Σ).

(* MAKERS
===================================================*)
  
  Lemma helper_queue (l : loc) (qs : gmapUR loc (dec_agreeR gname)) :
     l ↦ #() ∗ ([∗ map] l ↦ _ ∈ qs, ∃ γ, own γq (Link l γ) ∗ l ↦{½} #())
                   ⊢ ⌜( qs !! l = None)⌝.
  Proof.
    iIntros "(Hl & Hmap)".
    destruct (qs !! l) eqn:Heq; trivial.
    iDestruct (big_sepM_lookup _ qs l with "Hmap") as (v0') "(_ & Hl')"; first done.
    iDestruct (heap_mapsto_invalid l 1%Qp ½ _ _ with "[$Hl $Hl']") as %[].
    + auto.
  Qed.

  Lemma makeQueue_spec :
    queues_inv N γq 
    ⊢ WP makeQueue #() 
      {{ v, ∃ γ γ' w, ⌜v = queueAddrT_to_val w⌝ ∗ queue N γp γq w γ γ' [] }}.
  Proof.
    iIntros "Hqueues".
    wp_lam.
    wp_alloc qh as "Hqh". iDestruct "Hqh" as "(Hqh & Hqh')".
    wp_alloc qt as "Hqt".
    wp_bind (ref _)%E.
    iInv (n_inv_queues N) as (queues) ">(Hqa & Hmap)" "Hclose".
    wp_alloc qs as "Hqs".

    (* Alloc list for procs *)
    iDestruct (list_alloc []) as ">Hlist". iDestruct "Hlist" as (γ) "(Hlist & Hlist')".
    (* Alloc new inQueueAuth *)
    iMod (own_alloc (InQueueAuth [])) as (γ') "HinQa". done.
    (* Alloc new queue link with γ' *)
    iDestruct (helper_queue qs queues with "[$Hqs $Hmap]") as %Hlookup.

    iDestruct "Hqs" as "(Hqs & Hqs')".    

    iMod (own_update _ (LinkAll queues) 
                     ((LinkAll (<[ qs := DecAgree γ' ]> queues)) ⋅
                       (Link qs γ')) with "Hqa") as "Hqa".
    { rewrite /LinkAll /Link. 
      apply auth_update_alloc.
      apply alloc_singleton_local_update. auto. done. }

    rewrite own_op.
    iDestruct "Hqa" as "(Hqa & Hqf)".
    iDestruct (link_split with "Hqf") as "(Hqf & Hqf')".
    iMod ("Hclose" with "[Hmap Hqa Hqs Hqf]") as "_".
    { iNext. rewrite /invariants.queues.
      iExists (<[qs:=DecAgree γ']> queues).
      iFrame.
      iDestruct (big_sepM_insert _ queues qs with "[Hmap Hqs Hqf]") as "Hmap".
      + done.
      + iFrame. iExists γ'. iFrame.
      + iFrame. }
    iModIntro.
    wp_value.
    iExists γ,γ',{|qhead:=qh;qtail:=qt;qsent:=qs|}. iFrame. iSplitR=>//. 
    iExists NONEV,None,None. iFrame. iRight. iSplitR=>//.
    iFrame. rewrite /queue_cont. auto.
  Qed.

  Lemma helper_proc (p : procAddrT) v h :
    (pvall p) ↦ v ∗ ([∗ map] k↦x ∈ h, ∃ v0 : procValT,
                   ⌜x = (1%Qp, DecAgree v0)⌝ ∗ pvall k ↦{½} pvalv v0)
              ⊢ ⌜h !! p = None⌝.
  Proof.
    iIntros "(Hpv & Hmap)".
    destruct (h !! p) eqn:Heq; trivial.
    iDestruct (big_sepM_lookup _ h p with "Hmap") as (v0') "(_ & Hpv')"; first done.
    iDestruct (heap_mapsto_invalid with "[$Hpv $Hpv']") as %[]. auto. 
  Qed.
  
  Lemma makeProc_spec (v : val) :
    procs_inv N γp 
    ⊢ WP makeProc v
      {{ v', ∃ p, ⌜v' = procAddrT_to_val p⌝ ∗ qproc N γp γq p ∗ (pvall p) ↦{½} v }}.
  Proof.
    iIntros "#Hprocs".
    wp_let.
    wp_alloc pv as "Hpv".
    wp_alloc pq as "Hpq".
    wp_alloc pp as "Hpp".
    wp_bind (ref _)%E.
    iInv (n_inv_procs N) as (h) ">(Hpa & Hmap)" "Hclose".
    wp_alloc pn as "Hpn".

    set ( p := {| pvall:= pv;pqueuel:=pq;pprevl:=pp;pnextl:=pn|}).
    set ( pval := (ProcValT v None None None)).

    iDestruct (helper_proc p with "[Hpv Hmap]") as %Hlookup. iFrame.
    
    (* Alloc new proc in h *)
    iMod (own_update _ (ProcAuth h)
                     ((ProcAuth (<[ p := (1%Qp, DecAgree pval) ]> h)) ⋅
                     (Proc p 1%Qp pval)) with "Hpa") as "Hpa".
    { rewrite /ProcAuth /Proc.
      apply auth_update_alloc.
      apply alloc_singleton_local_update. auto. done. }
      
    rewrite own_op.
    iDestruct "Hpa" as "(Hpa & Hpf)".
    iDestruct "Hpv" as "(Hpv & Hpv')".
    
    iMod ("Hclose" with "[Hmap Hpa Hpv]") as "_".
    { iNext. rewrite /procs.
      iExists (<[p:=(1%Qp, DecAgree pval)]> h).
      iFrame.
      iDestruct (big_sepM_insert _ h p with "[Hmap Hpv]") as "Hmap".
      + done.
      + iFrame. iExists pval. by iFrame.
      + iFrame. }

    iDestruct (proc_own_split with "Hpf") as "(Hpf & Hpf')".

    iAssert (∃ v0 q0 p0 n0, proc γp γq p v0 q0 p0 n0)%I with "[-Hpf' Hpv']" as "HP".
    { iExists v, None, None, None. iFrame. }
    
    iMod (inv_alloc (n_inv_proc N p) ⊤ _ with "[HP]") as "#Hprocinv". iFrame.
    iModIntro. wp_value. iExists p. iSplitR=>//. iFrame. iExists v. iFrame; iFrame "#".
  Qed.

End makers.