From iris.heap_lang Require Import notation lib.assert.

(* PROGRAM 
===================================================*)

Definition unSOME : val :=
  λ: "p",
     match: "p" with NONE => assert #false | SOME "p'" => "p'" end.

(* Projections *)
Definition queue_head : val := λ: "p", Fst "p".
Definition queue_tail : val := λ: "p", Fst (Snd "p").
Definition queue_sent : val := λ: "p", Snd (Snd "p").

Definition pval : val := λ: "p", Fst "p".
Definition qref : val := λ: "p", Fst (Snd "p").
Definition prev : val := λ: "p", Fst (Snd (Snd "p")).
Definition next : val := λ: "p", Snd (Snd (Snd "p")).

Definition makeQueue : val :=
  λ: <>,
     (ref NONE, (ref NONE, ref #())).

Definition makeProc : val :=
  λ: "v",
     (ref "v", (ref NONE, (ref NONE, ref NONE))).  

Definition obtainLockEnq : val :=
  rec: "loop" "head" "sentinel" "h" :=
    let: "hv" := !"h" in
    if: ("hv" = SOME "sentinel") || (~ CAS "head" "hv" (SOME "sentinel"))
    then "h" <- !"head" ;; "loop" "head" "sentinel" "h"
    else #().
                              
Definition enqueue : val :=
  λ: "head" "tail" "s",
     λ: "p",
        let: "h" := ref !"head" in
           obtainLockEnq "head" "s" "h"
        ;; qref "p" <- SOME ("head",("tail","s")) 
        ;; match: !"h" with
              NONE => "tail" <- SOME "p"
                       ;; "head" <- SOME "p"
                       ;; #true
            | SOME "h'" =>  prev "p" <- !"tail"
                            ;; next (unSOME !"tail") <- SOME "p"
                            ;; "tail" <- SOME "p"
                            ;; "head" <- SOME "h'"
                            ;; #false
          end.

Definition obtainLockDeq : val :=
  rec: "loop" "head" "sentinel" "h" :=
    let: "hv" := !"h" in
    if: ("hv" = SOME "sentinel") || (~ CAS "head" "hv" (SOME "sentinel"))
    then "h" <- !"head" ;;
         if: (!"h") = NONE then #true
         else "loop" "head" "sentinel" "h"
    else #false.

Definition dequeue : val :=
  λ: "head" "tail" "s",
     λ: <>,
        let: "h" := ref !"head" in
        if: !"h" = NONE then NONE
        else let: "obtLock" := obtainLockDeq "head" "s" "h" in
             if: "obtLock" then NONE
             else let: "h'" := unSOME (!"h") in
                  let: "next" := !(next "h'") in
                  (if: "next" = NONE then
                     "tail" <- NONE
                   else prev (unSOME "next") <- NONE)
                    ;; next "h'" <- NONE
                    ;; qref "h'" <- NONE
                    ;; "head" <- "next"
                    ;; SOME "h'".

Definition tryDequeueEntry : val :=
  λ: "head" "tail" "s",
     λ: "p",
          let: "h" := ref !"head" in
          if: !"h" = NONE then #false
          else let: "obtLock" := obtainLockDeq "head" "s" "h" in
               if: "obtLock" then
                 #false
               else if: !(qref "p") = SOME ("head",("tail","s")) then
                      let: "next" := !(next "p") in
                      let: "prev" := !(prev "p") in
                      (if: "next" = NONE then
                         "tail" <- "prev"
                       else
                         prev (unSOME "next") <- "prev") 
                      ;; (if: "prev" = NONE then
                              "h" <- "next"
                          else
                            next (unSOME "prev") <- "next") 
                      ;; (prev "p") <- NONE 
                      ;; (next "p") <- NONE 
                      ;; (qref "p") <- NONE 
                      ;; "head" <- !"h" 
                      ;; #true
                    else
                      "head" <- !"h" ;; #false.
         
Opaque unSOME queue_head queue_tail queue_sent qref next prev 
       obtainLockEnq obtainLockDeq enqueue dequeue makeQueue 
       makeProc tryDequeueEntry.
