From iris.algebra Require Import gmap frac deprecated auth gset.
From iris.base_logic Require Import  big_op wsat invariants.
From iris.proofmode Require Import tactics.
From iris.heap_lang Require Import proofmode notation.
From stdpp Require Import mapset.
From fletch Require Import helpers definitions monoids invariants program.

Import dec_agree.

Section wp_helpers.

  Context {Σ : gFunctors} `{!heapG Σ, !fletchG Σ}
          (N : namespace) (γp γq γl : gname).
    
  Local Notation iProp := (iProp Σ).

(* SIMPLE SPECS
===================================================*)

  Lemma unSOME_spec p φ :
    φ (procAddrT_to_val p) ⊢ WP unSOME (option_procAddrT_to_val (Some p)) {{ φ }}.
  Proof.
    iIntros "H"; wp_lam; rewrite /option_procAddrT_to_val; by wp_match.
  Qed.

  Lemma queue_head_spec q φ :
    φ #(qhead q) ⊢ WP queue_head (queueAddrT_to_val q) {{ φ }}.
  Proof.
    iIntros "H"; destruct q; wp_lam; wp_proj; auto.
  Qed.

  Lemma queue_tail_spec q φ :
    φ #(qtail q) ⊢ WP queue_tail (queueAddrT_to_val q) {{ φ }}.
  Proof.
    iIntros "H"; destruct q; wp_lam; do 2 wp_proj; auto.
  Qed.

  Lemma queue_sent_spec q φ :
    φ #(qsent q) ⊢ WP queue_sent (queueAddrT_to_val q) {{ φ }}.
  Proof.
    iIntros "H"; destruct q; wp_lam; do 2 wp_proj; auto.
  Qed.

  Lemma pval_spec (p : procAddrT) φ :
    φ #(pvall p) ⊢ WP pval (procAddrT_to_val p) {{ φ }}.
  Proof. iIntros "H1"; destruct p; wp_lam; wp_proj; auto. Qed.
  
  Lemma qref_spec (p : procAddrT) φ :
    φ #(pqueuel p) ⊢ WP qref (procAddrT_to_val p) {{ φ }}.
  Proof. iIntros "H1"; destruct p; wp_lam; do 2 wp_proj; auto. Qed.

  Lemma prev_spec (p : procAddrT) φ :
    φ #(pprevl p) ⊢ WP prev (procAddrT_to_val p) {{ φ }}.
  Proof. iIntros "H1"; destruct p; wp_lam; do 3 wp_proj; auto. Qed.

  Lemma next_spec (p : procAddrT) φ :
    φ #(pnextl p) ⊢ WP next (procAddrT_to_val p) {{ φ }}.
  Proof. iIntros "H1"; destruct p; wp_lam; do 3 wp_proj; auto. Qed.
  
(* WP Helpers
===================================================*)

  Lemma wp_proc_load_qref (p : procAddrT) φ:
    proc_inv N γp γq p ∗ (∀ r, φ r)
    ⊢  WP ! (qref (procAddrT_to_val p)) {{ φ }}.
  Proof.
    iIntros "(#Hprocinv & Hpost)".
    wp_bind (qref _)%E.
    iApply qref_spec.
    iInv (n_inv_proc N p) as 
        (pvv' pqv' ppv' pnv') ">(Hpg & Hpql & Hprest)" "HcloseProcInv".
    wp_load. iMod ("HcloseProcInv" with "[Hpg Hpql Hprest]") as "_".
    { iNext. iExists pvv',pqv',ppv',pnv'; by iFrame. }
    iModIntro. iApply "Hpost".
  Qed.

  Lemma wp_proc_store_queue_none p pv q pp pn φ :
    procs_inv N γp ∗ proc_inv N γp γq p ∗
    own γp (Proc p ¾ (ProcValT pv (Some q) pp pn)) ∗
     ((∃ γ, own γp (Proc p ¾ (ProcValT pv None pp pn)) ∗ 
            own γq (Link (qsent q) γ) ∗ own γ (InQueue p)) -∗ φ #()) 
    ⊢ WP #(pqueuel p) <- NONE {{ φ }}.
  Proof.
    iIntros "(#Hprocsinv & #HprocInv & Hpg & Hvs)".
    iInv (n_inv_proc N p) as (pv0 pq0 pp0 pn0) ">Hproc" "HcloseProcInv".
    iDestruct "Hproc" as "(Hpg' & Hql & Hpl & Hnl & Hrest)".
    iDestruct (proc_own_eq with "[$Hpg $Hpg']") as %Heq. 
    inversion Heq; subst.
    wp_store.
    iDestruct (proc_own_full with "[$Hpg $Hpg']") as "(% & Hpg)".  
    iDestruct (proc_own_update N γp _ _ _ None pp pp pn pn (open_proc_mask N _)
               with "[$Hpg $Hprocsinv]")
      as "Hpgvs". rewrite /open_proc_mask. solve_ndisj.
    iMod "Hpgvs" as "Hpg". 
    iDestruct (proc_own_split with "Hpg") as "[Hpg Hpg']".
    iMod ("HcloseProcInv" with "[Hql Hpl Hnl Hpg]") as "_".
    iNext. iExists pv,None,pp,pn. iFrame. iModIntro. 
    iApply "Hvs". iFrame.
  Qed.

  Lemma wp_proc_store_queue_some p pv q pp pn φ :
    procs_inv N γp ∗ proc_inv N γp γq p ∗
    own γp (Proc p ¾ (ProcValT pv None pp pn)) ∗ 
    (∃ γ, own γq (Link (qsent q) γ) ∗ own γ (InQueue p)) ∗
    (own γp (Proc p ¾ (ProcValT pv (Some q) pp pn)) -∗ φ #()) 
    ⊢ WP #(pqueuel p) <- option_queueAddrT_to_val (Some q) {{ φ }}.
  Proof.
    iIntros "(#Hprocsinv & #HprocInv & Hpg & Hpqf & Hvs)".
    iInv (N .@ "proc" .@ p) as (pv0 pq0 pp0 pn0) ">Hproc" "HcloseProcInv".
    iDestruct "Hproc" as "(Hpg' & Hql & Hpl & Hnl & Hrest)".
    iDestruct (proc_own_eq with "[$Hpg $Hpg']") as %Heq.
    inversion Heq. subst.
    wp_store.
    iDestruct (proc_own_full with "[$Hpg $Hpg']") as "(% & Hpg)".
    iDestruct (proc_own_update N γp _ _ _ (Some q) pp pp pn pn (open_proc_mask N _)
               with "[$Hpg $Hprocsinv]") as "Hpgvs".
    { rewrite /open_proc_mask. solve_ndisj. }
    iMod "Hpgvs" as "Hpg". iDestruct (proc_own_split with "Hpg") as "[Hpg Hpg']".
    iMod ("HcloseProcInv" with "[Hql Hpl Hnl Hpg Hpqf]") as "_".
    iNext. iExists pv,(Some q),pp,pn. iFrame=>//.
    iModIntro. iApply "Hvs". iFrame.
  Qed.

  Lemma wp_proc_store_prev p pv pq pp pn pp' φ :
    procs_inv N γp ∗ proc_inv N γp γq p ∗
    own γp (Proc p ¾ (ProcValT pv pq pp pn)) ∗
    (own γp (Proc p ¾ (ProcValT pv pq pp' pn)) -∗ φ #())
    ⊢ WP #(pprevl p) <- option_procAddrT_to_val pp' {{ φ }}.
  Proof.
    iIntros "(#Hprocsinv & #HprocInv & Hpg & Hvs)".
    iInv (n_inv_proc N p) as (pv0 pq0 pp0 pn0) 
         ">(Hpg' & Hql & Hpl & Hnl & Hrest)" "HcloseProcInv".
    iDestruct (proc_own_eq with "[$Hpg $Hpg']") as %Heq. inversion Heq. subst.
    wp_store.
    iDestruct (proc_own_full with "[$Hpg $Hpg']") as "(% & Hpg)".
    iDestruct (proc_own_update N γp  _ _ _ _ pp pp' pn pn (open_proc_mask N _)
               with "[$Hpg $Hprocsinv]") as "Hpgvs".
    { rewrite /open_proc_mask. solve_ndisj. }
    iMod "Hpgvs" as "Hpg". iDestruct (proc_own_split with "Hpg") as "[Hpg Hpg']".
    iMod ("HcloseProcInv" with "[Hql Hpl Hnl Hpg Hrest]") as "_".
    iNext. iExists pv,pq,pp',pn. iFrame=>//. 
    iModIntro. iApply "Hvs". iFrame. 
  Qed.

  Lemma wp_proc_store_next p pv pq pp pn pn' φ :
    procs_inv N γp ∗ proc_inv N γp γq p ∗
    own γp (Proc p ¾ (ProcValT pv pq pp pn)) ∗
    (own γp (Proc p ¾ (ProcValT pv pq pp pn')) -∗ φ #())
    ⊢ WP #(pnextl p) <- option_procAddrT_to_val pn' {{ φ }}.
  Proof.
    iIntros "(#Hprocsinv & #HprocInv & Hpg & Hvs)".
    iInv (n_inv_proc N p) as (pv0 pq0 pp0 pn0) 
         ">(Hpg' & Hql & Hpl & Hnl & Hrest)" "HcloseProcInv".
    iDestruct (proc_own_eq with "[$Hpg $Hpg']") as %Heq; inversion Heq; subst.
    wp_store.
    iDestruct (proc_own_full with "[$Hpg $Hpg']") as "(% & Hpg)".
    iDestruct (proc_own_update N γp _ _ _ _ _ _ pn pn' (open_proc_mask N _)
               with "[$Hpg $Hprocsinv]") as "Hpgvs".
    { rewrite /open_proc_mask. solve_ndisj. }
    iMod "Hpgvs" as "Hpg". iDestruct (proc_own_split with "Hpg") as "[Hpg Hpg']".
    iMod ("HcloseProcInv" with "[Hql Hpl Hnl Hpg Hrest]") as "_".
    iNext. iExists pv,pq,pp,pn'. iFrame=>//. 
    iModIntro. iApply "Hvs". iFrame.
  Qed.

  Lemma wp_proc_load_prev_ghost (p : procAddrT) pv pq pp pn φ:
    proc_inv N γp γq p ∗ own γp (Proc p ¾ (ProcValT pv pq pp pn)) ∗ 
    (∀ r, ⌜r = (option_procAddrT_to_val pp)⌝ ∗ 
     own γp (Proc p ¾ (ProcValT pv pq pp pn)) -∗ φ r)
    ⊢  WP ! (prev (procAddrT_to_val p)) {{ φ }}.
  Proof.
    iIntros "(#Hprocinv & Hpg & Hpost)".
    wp_bind (prev _)%E. iApply prev_spec.
    iInv (n_inv_proc N p) as (pvv' pqv' ppv' pnv') 
          ">(Hpg' & Hpql & Hppl & Hpnl & Hprest)" "HcloseProcInv".
    iDestruct (proc_own_eq with "[$Hpg $Hpg']") as %Heq; inversion Heq; subst. 
    wp_load. iMod ("HcloseProcInv" with "[Hpql Hppl Hpnl Hpg' Hprest]") as "_".
    iNext. iExists pv,pq,pp,pn. iFrame. 
    iModIntro. iApply "Hpost"; iFrame=>//.
  Qed.

  Lemma wp_proc_load_next_ghost (p : procAddrT) pv pq pp pn φ:
    proc_inv N γp γq p ∗ own γp (Proc p ¾ (ProcValT pv pq pp pn)) ∗ 
    (∀ r, ⌜r = (option_procAddrT_to_val pn)⌝ ∗ 
     own γp (Proc p ¾ (ProcValT pv pq pp pn)) -∗ φ r)
    ⊢  WP ! (next (procAddrT_to_val p)) {{ φ }}.
  Proof.
    iIntros "(#Hprocinv & Hpg & Hpost)".
    wp_bind (next _)%E. iApply next_spec.
    iInv (n_inv_proc N p) as (pvv' pqv' ppv' pnv') 
       ">(Hpg' & Hpql & Hppl & Hpnl & Hprest)" "HcloseProcInv".
    iDestruct (proc_own_eq with "[$Hpg $Hpg']") as %Heq; inversion Heq; subst. 
    wp_load. iMod ("HcloseProcInv" with "[Hpql Hppl Hpnl Hpg' Hprest]") as "_".
    iNext. iExists pv,pq,pp,pn. by iFrame. iModIntro.
    iApply "Hpost"; iFrame=>//. 
  Qed.

  Lemma wp_proc_load_qref_ghost (p : procAddrT) pv pq pp pn φ:
    proc_inv N γp γq p ∗ own γp (Proc p ¾ (ProcValT pv pq pp pn)) ∗ 
    (∀ r, ⌜r = (option_queueAddrT_to_val pq)⌝ ∗ 
          own γp (Proc p ¾ (ProcValT pv pq pp pn)) -∗ φ r)
    ⊢  WP ! (qref (procAddrT_to_val p)) {{ φ }}.
  Proof.
    iIntros "(#Hprocinv & Hpg & Hpost)".
    wp_apply qref_spec.
    iInv (n_inv_proc N p) as (pvv' pqv' ppv' pnv') 
      ">(Hpg' & Hpql & Hppl & Hpnl & Hprest)" "HcloseProcInv".
    iDestruct (proc_own_eq with "[$Hpg $Hpg']") as %Heq; inversion Heq; subst. 
    wp_load. iMod ("HcloseProcInv" with "[Hpql Hppl Hpnl Hpg' Hprest]") as "_".
    iNext. iExists pv,pq,pp,pn. by iFrame. iModIntro.
    iApply "Hpost"; iFrame=>//. 
  Qed.

End wp_helpers.